#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import time
from weakref import ref as weak_ref_to
from typing import Any
from typing_extensions import Self
import numpy as np

import vedo.vtkclasses as vtki  # a wrapper for lazy imports

import vedo
from vedo import utils
from vedo.core import PointAlgorithms
from vedo.mesh import Mesh
from vedo.file_io import download
from vedo.visual import MeshVisual
from vedo.transformations import LinearTransform

__docformat__ = "google"

__doc__ = """
Work with tetrahedral meshes.

![](https://vedo.embl.es/images/volumetric/82767107-2631d500-9e25-11ea-967c-42558f98f721.jpg)
"""

__all__ = [
    "UnstructuredGrid",
    "TetMesh",
    "RectilinearGrid",
    "StructuredGrid",
]


#########################################################################
class UnstructuredGrid(PointAlgorithms, MeshVisual):
    """Support for UnstructuredGrid objects."""

    def __init__(self, inputobj=None):
        """
        Support for UnstructuredGrid objects.

        Arguments:
            inputobj : (list, vtkUnstructuredGrid, str)
                A list in the form `[points, cells, celltypes]`,
                or a vtkUnstructuredGrid object, or a filename

        Celltypes are identified by the following 
        [convention](https://vtk.org/doc/nightly/html/vtkCellType_8h_source.html).
        """
        super().__init__()

        self.dataset = None

        self.mapper = vtki.new("PolyDataMapper")
        self._actor = vtki.vtkActor()
        self._actor.retrieve_object = weak_ref_to(self)
        self._actor.SetMapper(self.mapper)
        self.properties = self._actor.GetProperty()

        self.transform = LinearTransform()
        self.point_locator = None
        self.cell_locator = None
        self.line_locator = None

        self.name = "UnstructuredGrid"
        self.filename = ""
        self.file_size = ""

        self.info = {}
        self.time = time.time()
        self.rendered_at = set()

        ######################################
        inputtype = str(type(inputobj))

        if inputobj is None:
            self.dataset = vtki.vtkUnstructuredGrid()

        elif utils.is_sequence(inputobj):

            pts, cells, celltypes = inputobj
            assert len(cells) == len(celltypes)

            self.dataset = vtki.vtkUnstructuredGrid()

            if not utils.is_sequence(cells[0]):
                tets = []
                nf = cells[0] + 1
                for i, cl in enumerate(cells):
                    if i in (nf, 0):
                        k = i + 1
                        nf = cl + k
                        cell = [cells[j + k] for j in range(cl)]
                        tets.append(cell)
                cells = tets

            # This would fill the points and use those to define orientation
            vpts = utils.numpy2vtk(pts, dtype=np.float32)
            points = vtki.vtkPoints()
            points.SetData(vpts)
            self.dataset.SetPoints(points)

            # Fill cells
            # https://vtk.org/doc/nightly/html/vtkCellType_8h_source.html
            for i, ct in enumerate(celltypes):
                if   ct == vtki.cell_types["VERTEX"]:
                    cell = vtki.vtkVertex()
                elif ct == vtki.cell_types["POLY_VERTEX"]:
                    cell = vtki.vtkPolyVertex()
                elif ct == vtki.cell_types["TETRA"]:
                    cell = vtki.vtkTetra()
                elif ct == vtki.cell_types["WEDGE"]:
                    cell = vtki.vtkWedge()
                elif ct == vtki.cell_types["LINE"]:
                    cell = vtki.vtkLine()
                elif ct == vtki.cell_types["POLY_LINE"]:
                    cell = vtki.vtkPolyLine()
                elif ct == vtki.cell_types["TRIANGLE"]:
                    cell = vtki.vtkTriangle()
                elif ct == vtki.cell_types["TRIANGLE_STRIP"]:
                    cell = vtki.vtkTriangleStrip()
                elif ct == vtki.cell_types["POLYGON"]:
                    cell = vtki.vtkPolygon()
                elif ct == vtki.cell_types["PIXEL"]:
                    cell = vtki.vtkPixel()
                elif ct == vtki.cell_types["QUAD"]:
                    cell = vtki.vtkQuad()
                elif ct == vtki.cell_types["VOXEL"]:
                    cell = vtki.vtkVoxel()
                elif ct == vtki.cell_types["PYRAMID"]:
                    cell = vtki.vtkPyramid()
                elif ct == vtki.cell_types["HEXAHEDRON"]:
                    cell = vtki.vtkHexahedron()
                elif ct == vtki.cell_types["HEXAGONAL_PRISM"]:
                    cell = vtki.vtkHexagonalPrism()
                elif ct == vtki.cell_types["PENTAGONAL_PRISM"]:
                    cell = vtki.vtkPentagonalPrism()
                elif ct == vtki.cell_types["QUADRATIC_TETRA"]:
                    from vtkmodules.vtkCommonDataModel import vtkQuadraticTetra
                    cell = vtkQuadraticTetra()
                elif ct == vtki.cell_types["QUADRATIC_HEXAHEDRON"]:
                    from vtkmodules.vtkCommonDataModel import vtkQuadraticHexahedron
                    cell = vtkQuadraticHexahedron()
                elif ct == vtki.cell_types["QUADRATIC_WEDGE"]:
                    from vtkmodules.vtkCommonDataModel import vtkQuadraticWedge
                    cell = vtkQuadraticWedge()
                elif ct == vtki.cell_types["QUADRATIC_PYRAMID"]:
                    from vtkmodules.vtkCommonDataModel import vtkQuadraticPyramid
                    cell = vtkQuadraticPyramid()
                elif ct == vtki.cell_types["QUADRATIC_LINEAR_QUAD"]:
                    from vtkmodules.vtkCommonDataModel import vtkQuadraticLinearQuad
                    cell = vtkQuadraticLinearQuad()
                elif ct == vtki.cell_types["QUADRATIC_LINEAR_WEDGE"]:
                    from vtkmodules.vtkCommonDataModel import vtkQuadraticLinearWedge
                    cell = vtkQuadraticLinearWedge()
                elif ct == vtki.cell_types["BIQUADRATIC_QUADRATIC_WEDGE"]:
                    from vtkmodules.vtkCommonDataModel import vtkBiQuadraticQuadraticWedge
                    cell = vtkBiQuadraticQuadraticWedge()
                elif ct == vtki.cell_types["BIQUADRATIC_QUADRATIC_HEXAHEDRON"]:
                    from vtkmodules.vtkCommonDataModel import vtkBiQuadraticQuadraticHexahedron
                    cell = vtkBiQuadraticQuadraticHexahedron()
                elif ct == vtki.cell_types["BIQUADRATIC_TRIANGLE"]:
                    from vtkmodules.vtkCommonDataModel import vtkBiQuadraticTriangle
                    cell = vtkBiQuadraticTriangle()
                elif ct == vtki.cell_types["CUBIC_LINE"]:
                    from vtkmodules.vtkCommonDataModel import vtkCubicLine
                    cell = vtkCubicLine()
                elif ct == vtki.cell_types["CONVEX_POINT_SET"]:
                    from vtkmodules.vtkCommonDataModel import vtkConvexPointSet
                    cell = vtkConvexPointSet()
                elif ct == vtki.cell_types["POLYHEDRON"]:
                    from vtkmodules.vtkCommonDataModel import vtkPolyhedron
                    cell = vtkPolyhedron()
                elif ct == vtki.cell_types["HIGHER_ORDER_TRIANGLE"]:
                    from vtkmodules.vtkCommonDataModel import vtkHigherOrderTriangle
                    cell = vtkHigherOrderTriangle()
                elif ct == vtki.cell_types["HIGHER_ORDER_QUAD"]:
                    from vtkmodules.vtkCommonDataModel import vtkHigherOrderQuadrilateral
                    cell = vtkHigherOrderQuadrilateral()
                elif ct == vtki.cell_types["HIGHER_ORDER_TETRAHEDRON"]:
                    from vtkmodules.vtkCommonDataModel import vtkHigherOrderTetra
                    cell = vtkHigherOrderTetra()
                elif ct == vtki.cell_types["HIGHER_ORDER_WEDGE"]:
                    from vtkmodules.vtkCommonDataModel import vtkHigherOrderWedge
                    cell = vtkHigherOrderWedge()
                elif ct == vtki.cell_types["HIGHER_ORDER_HEXAHEDRON"]:
                    from vtkmodules.vtkCommonDataModel import vtkHigherOrderHexahedron
                    cell = vtkHigherOrderHexahedron()
                elif ct == vtki.cell_types["LAGRANGE_CURVE"]:
                    from vtkmodules.vtkCommonDataModel import vtkLagrangeCurve
                    cell = vtkLagrangeCurve()
                elif ct == vtki.cell_types["LAGRANGE_TRIANGLE"]:
                    from vtkmodules.vtkCommonDataModel import vtkLagrangeTriangle
                    cell = vtkLagrangeTriangle()
                elif ct == vtki.cell_types["LAGRANGE_QUADRILATERAL"]:
                    from vtkmodules.vtkCommonDataModel import vtkLagrangeQuadrilateral
                    cell = vtkLagrangeQuadrilateral()
                elif ct == vtki.cell_types["LAGRANGE_TETRAHEDRON"]:
                    from vtkmodules.vtkCommonDataModel import vtkLagrangeTetra
                    cell = vtkLagrangeTetra()
                elif ct == vtki.cell_types["LAGRANGE_HEXAHEDRON"]:
                    from vtkmodules.vtkCommonDataModel import vtkLagrangeHexahedron
                    cell = vtkLagrangeHexahedron()
                elif ct == vtki.cell_types["LAGRANGE_WEDGE"]:
                    from vtkmodules.vtkCommonDataModel import vtkLagrangeWedge
                    cell = vtkLagrangeWedge()
                elif ct == vtki.cell_types["BEZIER_CURVE"]:
                    from vtkmodules.vtkCommonDataModel import vtkBezierCurve
                    cell = vtkBezierCurve()
                elif ct == vtki.cell_types["BEZIER_TRIANGLE"]:
                    from vtkmodules.vtkCommonDataModel import vtkBezierTriangle
                    cell = vtkBezierTriangle()
                elif ct == vtki.cell_types["BEZIER_QUADRILATERAL"]:
                    from vtkmodules.vtkCommonDataModel import vtkBezierQuadrilateral
                    cell = vtkBezierQuadrilateral()
                elif ct == vtki.cell_types["BEZIER_TETRAHEDRON"]:
                    from vtkmodules.vtkCommonDataModel import vtkBezierTetra
                    cell = vtkBezierTetra()
                elif ct == vtki.cell_types["BEZIER_HEXAHEDRON"]:
                    from vtkmodules.vtkCommonDataModel import vtkBezierHexahedron
                    cell = vtkBezierHexahedron()
                elif ct == vtki.cell_types["BEZIER_WEDGE"]:
                    from vtkmodules.vtkCommonDataModel import vtkBezierWedge
                    cell = vtkBezierWedge()
                else:
                    vedo.logger.error(
                        f"UnstructuredGrid: cell type {ct} not supported. Skip.")
                    continue

                cpids = cell.GetPointIds()
                cell_conn = cells[i]
                for j, pid in enumerate(cell_conn):
                    cpids.SetId(j, pid)
                self.dataset.InsertNextCell(ct, cpids)

        elif "UnstructuredGrid" in inputtype:
            self.dataset = inputobj

        elif isinstance(inputobj, str):
            if "https://" in inputobj:
                inputobj = download(inputobj, verbose=False)
            self.filename = inputobj
            if inputobj.endswith(".vtu"):
                reader = vtki.new("XMLUnstructuredGridReader")
            else:
                reader = vtki.new("UnstructuredGridReader")
            self.filename = inputobj
            reader.SetFileName(inputobj)
            reader.Update()
            self.dataset = reader.GetOutput()

        else:
            # this converts other types of vtk objects to UnstructuredGrid
            apf = vtki.new("AppendFilter")
            try:
                apf.AddInputData(inputobj)
            except TypeError:
                apf.AddInputData(inputobj.dataset)
            apf.Update()
            self.dataset = apf.GetOutput()

        self.properties.SetColor(0.89, 0.455, 0.671)  # pink7

        self.pipeline = utils.OperationNode(
            self, comment=f"#cells {self.dataset.GetNumberOfCells()}", c="#4cc9f0"
        )

    # ------------------------------------------------------------------
    def __str__(self):
        """Print a string summary of the `UnstructuredGrid` object."""
        module = self.__class__.__module__
        name = self.__class__.__name__
        out = vedo.printc(
            f"{module}.{name} at ({hex(self.memory_address())})".ljust(75),
            c="m", bold=True, invert=True, return_string=True,
        )
        out += "\x1b[0m\u001b[35m"

        out += "nr. of verts".ljust(14) + ": " + str(self.npoints) + "\n"
        out += "nr. of cells".ljust(14) + ": " + str(self.ncells)  + "\n"
        ct_arr = np.unique(self.cell_types_array)
        cnames = [k for k, v in vtki.cell_types.items() if v in ct_arr]
        out += "cell types".ljust(14) + ": " + str(cnames) + "\n"

        if self.npoints:
            out+="size".ljust(14)+ ": average=" + utils.precision(self.average_size(),6)
            out+=", diagonal="+ utils.precision(self.diagonal_size(), 6)+ "\n"
            out+="center of mass".ljust(14) + ": " + utils.precision(self.center_of_mass(),6)+"\n"

        bnds = self.bounds()
        bx1, bx2 = utils.precision(bnds[0], 3), utils.precision(bnds[1], 3)
        by1, by2 = utils.precision(bnds[2], 3), utils.precision(bnds[3], 3)
        bz1, bz2 = utils.precision(bnds[4], 3), utils.precision(bnds[5], 3)
        out += "bounds".ljust(14) + ":"
        out += " x=(" + bx1 + ", " + bx2 + "),"
        out += " y=(" + by1 + ", " + by2 + "),"
        out += " z=(" + bz1 + ", " + bz2 + ")\n"

        for key in self.pointdata.keys():
            arr = self.pointdata[key]
            rng = utils.precision(arr.min(), 3) + ", " + utils.precision(arr.max(), 3)
            mark_active = "pointdata"
            a_scalars = self.dataset.GetPointData().GetScalars()
            a_vectors = self.dataset.GetPointData().GetVectors()
            a_tensors = self.dataset.GetPointData().GetTensors()
            if a_scalars and a_scalars.GetName() == key:
                mark_active += " *"
            elif a_vectors and a_vectors.GetName() == key:
                mark_active += " **"
            elif a_tensors and a_tensors.GetName() == key:
                mark_active += " ***"
            out += mark_active.ljust(14) + f': "{key}" ({arr.dtype}), ndim={arr.ndim}'
            out += f", range=({rng})\n"

        for key in self.celldata.keys():
            arr = self.celldata[key]
            rng = utils.precision(arr.min(), 3) + ", " + utils.precision(arr.max(), 3)
            mark_active = "celldata"
            a_scalars = self.dataset.GetCellData().GetScalars()
            a_vectors = self.dataset.GetCellData().GetVectors()
            a_tensors = self.dataset.GetCellData().GetTensors()
            if a_scalars and a_scalars.GetName() == key:
                mark_active += " *"
            elif a_vectors and a_vectors.GetName() == key:
                mark_active += " **"
            elif a_tensors and a_tensors.GetName() == key:
                mark_active += " ***"
            out += mark_active.ljust(14) + f': "{key}" ({arr.dtype}), ndim={arr.ndim}'
            out += f", range=({rng})\n"

        for key in self.metadata.keys():
            arr = self.metadata[key]
            out += "metadata".ljust(14) + ": " + f'"{key}" ({len(arr)} values)\n'

        return out.rstrip() + "\x1b[0m"

    def _repr_html_(self):
        """
        HTML representation of the UnstructuredGrid object for Jupyter Notebooks.

        Returns:
            HTML text with the image and some properties.
        """
        import io
        import base64
        from PIL import Image

        library_name = "vedo.grids.UnstructuredGrid"
        help_url = "https://vedo.embl.es/docs/vedo/grids.html#UnstructuredGrid"

        arr = self.thumbnail()
        im = Image.fromarray(arr)
        buffered = io.BytesIO()
        im.save(buffered, format="PNG", quality=100)
        encoded = base64.b64encode(buffered.getvalue()).decode("utf-8")
        url = "data:image/png;base64," + encoded
        image = f"<img src='{url}'></img>"

        bounds = "<br/>".join(
            [
                utils.precision(min_x, 4) + " ... " + utils.precision(max_x, 4)
                for min_x, max_x in zip(self.bounds()[::2], self.bounds()[1::2])
            ]
        )

        help_text = ""
        if self.name:
            help_text += f"<b> {self.name}: &nbsp&nbsp</b>"
        help_text += '<b><a href="' + help_url + '" target="_blank">' + library_name + "</a></b>"
        if self.filename:
            dots = ""
            if len(self.filename) > 30:
                dots = "..."
            help_text += f"<br/><code><i>({dots}{self.filename[-30:]})</i></code>"

        pdata = ""
        if self.dataset.GetPointData().GetScalars():
            if self.dataset.GetPointData().GetScalars().GetName():
                name = self.dataset.GetPointData().GetScalars().GetName()
                pdata = "<tr><td><b> point data array </b></td><td>" + name + "</td></tr>"

        cdata = ""
        if self.dataset.GetCellData().GetScalars():
            if self.dataset.GetCellData().GetScalars().GetName():
                name = self.dataset.GetCellData().GetScalars().GetName()
                cdata = "<tr><td><b> cell data array </b></td><td>" + name + "</td></tr>"

        pts = self.coordinates
        cm = np.mean(pts, axis=0)

        all = [
            "<table>",
            "<tr>",
            "<td>", image, "</td>",
            "<td style='text-align: center; vertical-align: center;'><br/>", help_text,
            "<table>",
            "<tr><td><b> bounds </b> <br/> (x/y/z) </td><td>" + str(bounds) + "</td></tr>",
            "<tr><td><b> center of mass </b></td><td>" + utils.precision(cm,3) + "</td></tr>",
            # "<tr><td><b> average size </b></td><td>" + str(average_size) + "</td></tr>",
            "<tr><td><b> nr. points&nbsp/&nbspcells </b></td><td>"
            + str(self.npoints) + "&nbsp/&nbsp" + str(self.ncells) + "</td></tr>",
            pdata,
            cdata,
            "</table>",
            "</table>",
        ]
        return "\n".join(all)

    @property
    def actor(self):
        """Return the `vtkActor` of the object."""
        # print("building actor")
        gf = vtki.new("GeometryFilter")
        gf.SetInputData(self.dataset)
        gf.Update()
        out = gf.GetOutput()
        self.mapper.SetInputData(out)
        self.mapper.Modified()
        return self._actor

    @actor.setter
    def actor(self, _):
        pass

    def _update(self, data, reset_locators=False):
        self.dataset = data
        if reset_locators:
            self.cell_locator = None
            self.point_locator = None
        return self
    
    def merge(self, *others) -> Self:
        """
        Merge multiple datasets into one single `UnstrcturedGrid`.
        """
        apf = vtki.new("AppendFilter")
        for o in others:
            if isinstance(o, UnstructuredGrid):
                apf.AddInputData(o.dataset)
            elif isinstance(o, vtki.vtkUnstructuredGrid):
                apf.AddInputData(o)
            else:
                vedo.printc("Error: cannot merge type", type(o), c="r")
        apf.Update()
        self._update(apf.GetOutput())
        self.pipeline = utils.OperationNode(
            "merge", parents=[self, *others], c="#9e2a2b"
        )
        return self

    def copy(self, deep=True) -> "UnstructuredGrid":
        """Return a copy of the object. Alias of `clone()`."""
        return self.clone(deep=deep)

    def clone(self, deep=True) -> "UnstructuredGrid":
        """Clone the UnstructuredGrid object to yield an exact copy."""
        ug = vtki.vtkUnstructuredGrid()
        if deep:
            ug.DeepCopy(self.dataset)
        else:
            ug.ShallowCopy(self.dataset)
        if isinstance(self, vedo.UnstructuredGrid):
            cloned = vedo.UnstructuredGrid(ug)
        else:
            cloned = vedo.TetMesh(ug)

        cloned.copy_properties_from(self)

        cloned.pipeline = utils.OperationNode(
            "clone", parents=[self], shape="diamond", c="#bbe1ed"
        )
        return cloned

    def bounds(self) -> np.ndarray:
        """
        Get the object bounds.
        Returns a list in format `[xmin,xmax, ymin,ymax, zmin,zmax]`.
        """
        # OVERRIDE CommonAlgorithms.bounds() which is too slow
        return np.array(self.dataset.GetBounds())

    def threshold(self, name=None, above=None, below=None, on="cells") -> Self:
        """
        Threshold the tetrahedral mesh by a cell scalar value.
        Reduce to only tets which satisfy the threshold limits.

        - if `above = below` will only select tets with that specific value.
        - if `above > below` selection range is flipped.

        Set keyword "on" to either "cells" or "points".
        """
        th = vtki.new("Threshold")
        th.SetInputData(self.dataset)

        if name is None:
            if self.celldata.keys():
                name = self.celldata.keys()[0]
                th.SetInputArrayToProcess(0, 0, 0, 1, name)
            elif self.pointdata.keys():
                name = self.pointdata.keys()[0]
                th.SetInputArrayToProcess(0, 0, 0, 0, name)
            if name is None:
                vedo.logger.warning("cannot find active array. Skip.")
                return self
        else:
            if on.startswith("c"):
                th.SetInputArrayToProcess(0, 0, 0, 1, name)
            else:
                th.SetInputArrayToProcess(0, 0, 0, 0, name)

        if above is not None:
            th.SetLowerThreshold(above)

        if below is not None:
            th.SetUpperThreshold(below)

        th.Update()
        return self._update(th.GetOutput())

    def isosurface(self, value=None, flying_edges=False) -> "vedo.mesh.Mesh":
        """
        Return an `Mesh` isosurface extracted from the `Volume` object.

        Set `value` as single float or list of values to draw the isosurface(s).
        Use flying_edges for faster results (but sometimes can interfere with `smooth()`).

        Examples:
            - [isosurfaces1.py](https://github.com/marcomusy/vedo/tree/master/examples/volumetric/isosurfaces1.py)

                ![](https://vedo.embl.es/images/volumetric/isosurfaces.png)
        """
        scrange = self.dataset.GetScalarRange()

        if flying_edges:
            cf = vtki.new("FlyingEdges3D")
            cf.InterpolateAttributesOn()
        else:
            cf = vtki.new("ContourFilter")
            cf.UseScalarTreeOn()

        cf.SetInputData(self.dataset)
        cf.ComputeNormalsOn()

        if utils.is_sequence(value):
            cf.SetNumberOfContours(len(value))
            for i, t in enumerate(value):
                cf.SetValue(i, t)
        else:
            if value is None:
                value = (2 * scrange[0] + scrange[1]) / 3.0
                # print("automatic isosurface value =", value)
            cf.SetValue(0, value)

        cf.Update()
        poly = cf.GetOutput()

        out = vedo.mesh.Mesh(poly, c=None).flat()
        out.mapper.SetScalarRange(scrange[0], scrange[1])

        out.pipeline = utils.OperationNode(
            "isosurface",
            parents=[self],
            comment=f"#pts {out.dataset.GetNumberOfPoints()}",
            c="#4cc9f0:#e9c46a",
        )
        return out

    def shrink(self, fraction=0.8) -> Self:
        """
        Shrink the individual cells.

        ![](https://vedo.embl.es/images/feats/shrink_hex.png)
        """
        sf = vtki.new("ShrinkFilter")
        sf.SetInputData(self.dataset)
        sf.SetShrinkFactor(fraction)
        sf.Update()
        out = sf.GetOutput()
        self._update(out)
        self.pipeline = utils.OperationNode(
            "shrink", comment=f"by {fraction}", parents=[self], c="#9e2a2b"
        )
        return self

    def tomesh(self, fill=False, shrink=1.0) -> "vedo.mesh.Mesh":
        """
        Build a polygonal `Mesh` from the current object.

        If `fill=True`, the interior faces of all the cells are created.
        (setting a `shrink` value slightly smaller than the default 1.0
        can avoid flickering due to internal adjacent faces).

        If `fill=False`, only the boundary faces will be generated.
        """
        gf = vtki.new("GeometryFilter")
        if fill:
            sf = vtki.new("ShrinkFilter")
            sf.SetInputData(self.dataset)
            sf.SetShrinkFactor(shrink)
            sf.Update()
            gf.SetInputData(sf.GetOutput())
            gf.Update()
            poly = gf.GetOutput()
        else:
            gf.SetInputData(self.dataset)
            gf.Update()
            poly = gf.GetOutput()

        msh = vedo.mesh.Mesh(poly)
        msh.copy_properties_from(self)
        msh.pipeline = utils.OperationNode(
            "tomesh", parents=[self], comment=f"fill={fill}", c="#9e2a2b:#e9c46a"
        )
        return msh

    @property
    def cell_types_array(self):
        """Return the list of cell types in the dataset."""
        uarr = self.dataset.GetCellTypesArray()
        return utils.vtk2numpy(uarr)

    def extract_cells_by_type(self, ctype) -> "UnstructuredGrid":
        """Extract a specific cell type and return a new `UnstructuredGrid`."""
        if isinstance(ctype, str):
            try:
                ctype = vtki.cell_types[ctype.upper()]
            except KeyError:
                vedo.logger.error(f"extract_cells_by_type: cell type {ctype} does not exist. Skip.")
                return self
        uarr = self.dataset.GetCellTypesArray()
        ctarrtyp = np.where(utils.vtk2numpy(uarr) == ctype)[0]
        uarrtyp = utils.numpy2vtk(ctarrtyp, deep=False, dtype="id")
        selection_node = vtki.new("SelectionNode")
        selection_node.SetFieldType(vtki.get_class("SelectionNode").CELL)
        selection_node.SetContentType(vtki.get_class("SelectionNode").INDICES)
        selection_node.SetSelectionList(uarrtyp)
        selection = vtki.new("Selection")
        selection.AddNode(selection_node)
        es = vtki.new("ExtractSelection")
        es.SetInputData(0, self.dataset)
        es.SetInputData(1, selection)
        es.Update()

        ug = UnstructuredGrid(es.GetOutput())
        ug.pipeline = utils.OperationNode(
            "extract_cell_type", comment=f"type {ctype}", c="#edabab", parents=[self]
        )
        return ug

    def extract_cells_by_id(self, idlist, use_point_ids=False) -> "UnstructuredGrid":
        """Return a new `UnstructuredGrid` composed of the specified subset of indices."""
        selection_node = vtki.new("SelectionNode")
        if use_point_ids:
            selection_node.SetFieldType(vtki.get_class("SelectionNode").POINT)
            contcells = vtki.get_class("SelectionNode").CONTAINING_CELLS()
            selection_node.GetProperties().Set(contcells, 1)
        else:
            selection_node.SetFieldType(vtki.get_class("SelectionNode").CELL)
        selection_node.SetContentType(vtki.get_class("SelectionNode").INDICES)
        vidlist = utils.numpy2vtk(idlist, dtype="id")
        selection_node.SetSelectionList(vidlist)
        selection = vtki.new("Selection")
        selection.AddNode(selection_node)
        es = vtki.new("ExtractSelection")
        es.SetInputData(0, self)
        es.SetInputData(1, selection)
        es.Update()

        ug = UnstructuredGrid(es.GetOutput())
        pr = vtki.vtkProperty()
        pr.DeepCopy(self.properties)
        ug.actor.SetProperty(pr)
        ug.properties = pr

        ug.mapper.SetLookupTable(utils.ctf2lut(self))
        ug.pipeline = utils.OperationNode(
            "extract_cells_by_id",
            parents=[self],
            comment=f"#cells {self.dataset.GetNumberOfCells()}",
            c="#9e2a2b",
        )
        return ug

    def find_cell(self, p: list) -> int:
        """Locate the cell that contains a point and return the cell ID."""
        if self.cell_locator is None:
            self.cell_locator = vtki.new("CellLocator")
            self.cell_locator.SetDataSet(self.dataset)
            self.cell_locator.BuildLocator()
        cid = self.cell_locator.FindCell(p)
        return cid

    def clean(self) -> Self:
        """
        Cleanup unused points and empty cells
        """
        cl = vtki.new("StaticCleanUnstructuredGrid")
        cl.SetInputData(self.dataset)
        cl.RemoveUnusedPointsOn()
        cl.ProduceMergeMapOff()
        cl.AveragePointDataOff()
        cl.Update()

        self._update(cl.GetOutput())
        self.pipeline = utils.OperationNode(
            "clean",
            parents=[self],
            comment=f"#cells {self.dataset.GetNumberOfCells()}",
            c="#9e2a2b",
        )
        return self

    def extract_cells_on_plane(self, origin: tuple, normal: tuple) -> Self:
        """
        Extract cells that are lying of the specified surface.
        """
        bf = vtki.new("3DLinearGridCrinkleExtractor")
        bf.SetInputData(self.dataset)
        bf.CopyPointDataOn()
        bf.CopyCellDataOn()
        bf.RemoveUnusedPointsOff()

        plane = vtki.new("Plane")
        plane.SetOrigin(origin)
        plane.SetNormal(normal)
        bf.SetImplicitFunction(plane)
        bf.Update()

        self._update(bf.GetOutput(), reset_locators=False)
        self.pipeline = utils.OperationNode(
            "extract_cells_on_plane",
            parents=[self],
            comment=f"#cells {self.dataset.GetNumberOfCells()}",
            c="#9e2a2b",
        )
        return self

    def extract_cells_on_sphere(self, center: tuple, radius: tuple) -> Self:
        """
        Extract cells that are lying of the specified surface.
        """
        bf = vtki.new("3DLinearGridCrinkleExtractor")
        bf.SetInputData(self.dataset)
        bf.CopyPointDataOn()
        bf.CopyCellDataOn()
        bf.RemoveUnusedPointsOff()

        sph = vtki.new("Sphere")
        sph.SetRadius(radius)
        sph.SetCenter(center)
        bf.SetImplicitFunction(sph)
        bf.Update()

        self._update(bf.GetOutput())
        self.pipeline = utils.OperationNode(
            "extract_cells_on_sphere",
            parents=[self],
            comment=f"#cells {self.dataset.GetNumberOfCells()}",
            c="#9e2a2b",
        )
        return self

    def extract_cells_on_cylinder(self, center: tuple, axis: tuple, radius: float) -> Self:
        """
        Extract cells that are lying of the specified surface.
        """
        bf = vtki.new("3DLinearGridCrinkleExtractor")
        bf.SetInputData(self.dataset)
        bf.CopyPointDataOn()
        bf.CopyCellDataOn()
        bf.RemoveUnusedPointsOff()

        cyl = vtki.new("Cylinder")
        cyl.SetRadius(radius)
        cyl.SetCenter(center)
        cyl.SetAxis(axis)
        bf.SetImplicitFunction(cyl)
        bf.Update()

        self.pipeline = utils.OperationNode(
            "extract_cells_on_cylinder",
            parents=[self],
            comment=f"#cells {self.dataset.GetNumberOfCells()}",
            c="#9e2a2b",
        )
        self._update(bf.GetOutput())
        return self

    def cut_with_plane(self, origin=(0, 0, 0), normal="x") -> "UnstructuredGrid":
        """
        Cut the object with the plane defined by a point and a normal.

        Arguments:
            origin : (list)
                the cutting plane goes through this point
            normal : (list, str)
                normal vector to the cutting plane
        """
        # if isinstance(self, vedo.Volume):
        #     raise RuntimeError("cut_with_plane() is not applicable to Volume objects.")

        strn = str(normal)
        if strn   ==  "x": normal = (1, 0, 0)
        elif strn ==  "y": normal = (0, 1, 0)
        elif strn ==  "z": normal = (0, 0, 1)
        elif strn == "-x": normal = (-1, 0, 0)
        elif strn == "-y": normal = (0, -1, 0)
        elif strn == "-z": normal = (0, 0, -1)
        plane = vtki.new("Plane")
        plane.SetOrigin(origin)
        plane.SetNormal(normal)
        clipper = vtki.new("ClipDataSet")
        clipper.SetInputData(self.dataset)
        clipper.SetClipFunction(plane)
        clipper.GenerateClipScalarsOff()
        clipper.GenerateClippedOutputOff()
        clipper.SetValue(0)
        clipper.Update()
        cout = clipper.GetOutput()

        if isinstance(cout, vtki.vtkUnstructuredGrid):
            ug = vedo.UnstructuredGrid(cout)
            if isinstance(self, vedo.UnstructuredGrid):
                self._update(cout)
                self.pipeline = utils.OperationNode("cut_with_plane", parents=[self], c="#9e2a2b")
                return self
            ug.pipeline = utils.OperationNode("cut_with_plane", parents=[self], c="#9e2a2b")
            return ug

        else:
            self._update(cout)
            self.pipeline = utils.OperationNode("cut_with_plane", parents=[self], c="#9e2a2b")
            return self

    def cut_with_box(self, box: Any) -> "UnstructuredGrid":
        """
        Cut the grid with the specified bounding box.

        Parameter box has format [xmin, xmax, ymin, ymax, zmin, zmax].
        If an object is passed, its bounding box are used.

        This method always returns a TetMesh object.

        Example:
        ```python
        from vedo import *
        tmesh = TetMesh(dataurl+'limb_ugrid.vtk')
        tmesh.color('rainbow')
        cu = Cube(side=500).x(500) # any Mesh works
        tmesh.cut_with_box(cu).show(axes=1)
        ```

        ![](https://vedo.embl.es/images/feats/tet_cut_box.png)
        """
        bc = vtki.new("BoxClipDataSet")
        bc.SetInputData(self.dataset)
        try:
            boxb = box.bounds()
        except AttributeError:
            boxb = box

        bc.SetBoxClip(*boxb)
        bc.Update()
        cout = bc.GetOutput()

        # output of vtkBoxClipDataSet is always tetrahedrons
        tm = vedo.TetMesh(cout)
        tm.pipeline = utils.OperationNode("cut_with_box", parents=[self], c="#9e2a2b")
        return tm

    def cut_with_mesh(self, mesh: Mesh, invert=False, whole_cells=False, on_boundary=False) -> "UnstructuredGrid":
        """
        Cut a `UnstructuredGrid` or `TetMesh` with a `Mesh`.

        Use `invert` to return cut off part of the input object.
        """
        ug = self.dataset

        ippd = vtki.new("ImplicitPolyDataDistance")
        ippd.SetInput(mesh.dataset)

        if whole_cells or on_boundary:
            clipper = vtki.new("ExtractGeometry")
            clipper.SetInputData(ug)
            clipper.SetImplicitFunction(ippd)
            clipper.SetExtractInside(not invert)
            clipper.SetExtractBoundaryCells(False)
            if on_boundary:
                clipper.SetExtractBoundaryCells(True)
                clipper.SetExtractOnlyBoundaryCells(True)
        else:
            signed_dists = vtki.vtkFloatArray()
            signed_dists.SetNumberOfComponents(1)
            signed_dists.SetName("SignedDistance")
            for pointId in range(ug.GetNumberOfPoints()):
                p = ug.GetPoint(pointId)
                signed_dist = ippd.EvaluateFunction(p)
                signed_dists.InsertNextValue(signed_dist)
            ug.GetPointData().AddArray(signed_dists)
            ug.GetPointData().SetActiveScalars("SignedDistance")  # NEEDED
            clipper = vtki.new("ClipDataSet")
            clipper.SetInputData(ug)
            clipper.SetInsideOut(not invert)
            clipper.SetValue(0.0)

        clipper.Update()

        out = vedo.UnstructuredGrid(clipper.GetOutput())
        out.pipeline = utils.OperationNode("cut_with_mesh", parents=[self], c="#9e2a2b")
        return out


##########################################################################
class TetMesh(UnstructuredGrid):
    """The class describing tetrahedral meshes."""

    def __init__(self, inputobj=None):
        """
        Arguments:
            inputobj : (vtkUnstructuredGrid, list, str, tetgenpy.TetgenIO)
                list of points and tet indices, or filename
        """
        super().__init__()

        self.dataset = None

        self.mapper = vtki.new("PolyDataMapper")
        self._actor = vtki.vtkActor()
        self._actor.retrieve_object = weak_ref_to(self)
        self._actor.SetMapper(self.mapper)
        self.properties = self._actor.GetProperty()

        self.name = "TetMesh"

        # print('TetMesh inputtype', type(inputobj))

        ###################
        if inputobj is None:
            self.dataset = vtki.vtkUnstructuredGrid()

        elif isinstance(inputobj, vtki.vtkUnstructuredGrid):
            self.dataset = inputobj

        elif isinstance(inputobj, UnstructuredGrid):
            self.dataset = inputobj.dataset

        elif "TetgenIO" in str(type(inputobj)):  # tetgenpy object
            inputobj = [inputobj.points(), inputobj.tetrahedra()]

        elif isinstance(inputobj, vtki.vtkRectilinearGrid):
            r2t = vtki.new("RectilinearGridToTetrahedra")
            r2t.SetInputData(inputobj)
            r2t.RememberVoxelIdOn()
            r2t.SetTetraPerCellTo6()
            r2t.Update()
            self.dataset = r2t.GetOutput()

        elif isinstance(inputobj, vtki.vtkDataSet):
            r2t = vtki.new("DataSetTriangleFilter")
            r2t.SetInputData(inputobj)
            r2t.TetrahedraOnlyOn()
            r2t.Update()
            self.dataset = r2t.GetOutput()

        elif isinstance(inputobj, str):
            if "https://" in inputobj:
                inputobj = download(inputobj, verbose=False)
            if inputobj.endswith(".vtu"):
                reader = vtki.new("XMLUnstructuredGridReader")
            else:
                reader = vtki.new("UnstructuredGridReader")

            if not os.path.isfile(inputobj):
                # for some reason vtk Reader does not complain
                vedo.logger.error(f"file {inputobj} not found")
                raise FileNotFoundError

            self.filename = inputobj
            reader.SetFileName(inputobj)
            reader.Update()
            ug = reader.GetOutput()

            tt = vtki.new("DataSetTriangleFilter")
            tt.SetInputData(ug)
            tt.SetTetrahedraOnly(True)
            tt.Update()
            self.dataset = tt.GetOutput()

        ###############################
        if utils.is_sequence(inputobj):
            self.dataset = vtki.vtkUnstructuredGrid()

            points, cells = inputobj
            if len(points) == 0:
                return
            if not utils.is_sequence(points[0]):
                return
            if len(cells) == 0:
                return

            if not utils.is_sequence(cells[0]):
                tets = []
                nf = cells[0] + 1
                for i, cl in enumerate(cells):
                    if i in (nf, 0):
                        k = i + 1
                        nf = cl + k
                        cell = [cells[j + k] for j in range(cl)]
                        tets.append(cell)
                cells = tets

            source_points = vtki.vtkPoints()
            varr = utils.numpy2vtk(points, dtype=np.float32)
            source_points.SetData(varr)
            self.dataset.SetPoints(source_points)

            source_tets = vtki.vtkCellArray()
            for f in cells:
                ele = vtki.vtkTetra()
                pid = ele.GetPointIds()
                for i, fi in enumerate(f):
                    pid.SetId(i, fi)
                source_tets.InsertNextCell(ele)
            self.dataset.SetCells(vtki.cell_types["TETRA"], source_tets)

        if not self.dataset:
            vedo.logger.error(f"cannot understand input type {type(inputobj)}")
            return

        self.properties.SetColor(0.352, 0.612, 0.996)  # blue7

        self.pipeline = utils.OperationNode(
            self, comment=f"#tets {self.dataset.GetNumberOfCells()}", c="#9e2a2b"
        )

    ##################################################################
    def __str__(self):
        """Print a string summary of the `TetMesh` object."""
        module = self.__class__.__module__
        name = self.__class__.__name__
        out = vedo.printc(
            f"{module}.{name} at ({hex(self.memory_address())})".ljust(75),
            c="c", bold=True, invert=True, return_string=True,
        )
        out += "\x1b[0m\u001b[36m"

        out += "nr. of verts".ljust(14) + ": " + str(self.npoints) + "\n"
        out += "nr. of tetras".ljust(14) + ": " + str(self.ncells) + "\n"

        if self.npoints:
            out+="size".ljust(14)+ ": average=" + utils.precision(self.average_size(),6)
            out+=", diagonal="+ utils.precision(self.diagonal_size(), 6)+ "\n"
            out+="center of mass".ljust(14) + ": " + utils.precision(self.center_of_mass(),6)+"\n"

        bnds = self.bounds()
        bx1, bx2 = utils.precision(bnds[0], 3), utils.precision(bnds[1], 3)
        by1, by2 = utils.precision(bnds[2], 3), utils.precision(bnds[3], 3)
        bz1, bz2 = utils.precision(bnds[4], 3), utils.precision(bnds[5], 3)
        out += "bounds".ljust(14) + ":"
        out += " x=(" + bx1 + ", " + bx2 + "),"
        out += " y=(" + by1 + ", " + by2 + "),"
        out += " z=(" + bz1 + ", " + bz2 + ")\n"

        for key in self.pointdata.keys():
            arr = self.pointdata[key]
            rng = utils.precision(arr.min(), 3) + ", " + utils.precision(arr.max(), 3)
            mark_active = "pointdata"
            a_scalars = self.dataset.GetPointData().GetScalars()
            a_vectors = self.dataset.GetPointData().GetVectors()
            a_tensors = self.dataset.GetPointData().GetTensors()
            if a_scalars and a_scalars.GetName() == key:
                mark_active += " *"
            elif a_vectors and a_vectors.GetName() == key:
                mark_active += " **"
            elif a_tensors and a_tensors.GetName() == key:
                mark_active += " ***"
            out += mark_active.ljust(14) + f': "{key}" ({arr.dtype}), ndim={arr.ndim}'
            out += f", range=({rng})\n"

        for key in self.celldata.keys():
            arr = self.celldata[key]
            rng = utils.precision(arr.min(), 3) + ", " + utils.precision(arr.max(), 3)
            mark_active = "celldata"
            a_scalars = self.dataset.GetCellData().GetScalars()
            a_vectors = self.dataset.GetCellData().GetVectors()
            a_tensors = self.dataset.GetCellData().GetTensors()
            if a_scalars and a_scalars.GetName() == key:
                mark_active += " *"
            elif a_vectors and a_vectors.GetName() == key:
                mark_active += " **"
            elif a_tensors and a_tensors.GetName() == key:
                mark_active += " ***"
            out += mark_active.ljust(14) + f': "{key}" ({arr.dtype}), ndim={arr.ndim}'
            out += f", range=({rng})\n"

        for key in self.metadata.keys():
            arr = self.metadata[key]
            out += "metadata".ljust(14) + ": " + f'"{key}" ({len(arr)} values)\n'

        return out.rstrip() + "\x1b[0m"

    def _repr_html_(self):
        """
        HTML representation of the TetMesh object for Jupyter Notebooks.

        Returns:
            HTML text with the image and some properties.
        """
        import io
        import base64
        from PIL import Image

        library_name = "vedo.grids.TetMesh"
        help_url = "https://vedo.embl.es/docs/vedo/grids.html#TetMesh"

        arr = self.thumbnail()
        im = Image.fromarray(arr)
        buffered = io.BytesIO()
        im.save(buffered, format="PNG", quality=100)
        encoded = base64.b64encode(buffered.getvalue()).decode("utf-8")
        url = "data:image/png;base64," + encoded
        image = f"<img src='{url}'></img>"

        bounds = "<br/>".join(
            [
                utils.precision(min_x, 4) + " ... " + utils.precision(max_x, 4)
                for min_x, max_x in zip(self.bounds()[::2], self.bounds()[1::2])
            ]
        )

        help_text = ""
        if self.name:
            help_text += f"<b> {self.name}: &nbsp&nbsp</b>"
        help_text += '<b><a href="' + help_url + '" target="_blank">' + library_name + "</a></b>"
        if self.filename:
            dots = ""
            if len(self.filename) > 30:
                dots = "..."
            help_text += f"<br/><code><i>({dots}{self.filename[-30:]})</i></code>"

        pdata = ""
        if self.dataset.GetPointData().GetScalars():
            if self.dataset.GetPointData().GetScalars().GetName():
                name = self.dataset.GetPointData().GetScalars().GetName()
                pdata = "<tr><td><b> point data array </b></td><td>" + name + "</td></tr>"

        cdata = ""
        if self.dataset.GetCellData().GetScalars():
            if self.dataset.GetCellData().GetScalars().GetName():
                name = self.dataset.GetCellData().GetScalars().GetName()
                cdata = "<tr><td><b> cell data array </b></td><td>" + name + "</td></tr>"

        pts = self.coordinates
        cm = np.mean(pts, axis=0)

        allt = [
            "<table>",
            "<tr>",
            "<td>", image, "</td>",
            "<td style='text-align: center; vertical-align: center;'><br/>", help_text,
            "<table>",
            "<tr><td><b> bounds </b> <br/> (x/y/z) </td><td>" + str(bounds) + "</td></tr>",
            "<tr><td><b> center of mass </b></td><td>" + utils.precision(cm,3) + "</td></tr>",
            "<tr><td><b> nr. points&nbsp/&nbsptets </b></td><td>"
            + str(self.npoints) + "&nbsp/&nbsp" + str(self.ncells) + "</td></tr>",
            pdata,
            cdata,
            "</table>",
            "</table>",
        ]
        return "\n".join(allt)

    def compute_quality(self, metric=7) -> np.ndarray:
        """
        Calculate functions of quality for the elements of a tetrahedral mesh.
        This method adds to the mesh a cell array named "Quality".

        Arguments:
            metric : (int)
                type of estimators:
                    - EDGE RATIO, 0
                    - ASPECT RATIO, 1
                    - RADIUS RATIO, 2
                    - ASPECT FROBENIUS, 3
                    - MIN_ANGLE, 4
                    - COLLAPSE RATIO, 5
                    - ASPECT GAMMA, 6
                    - VOLUME, 7
                    - ...

        See class [vtkMeshQuality](https://vtk.org/doc/nightly/html/classvtkMeshQuality.html)
        for an explanation of the meaning of each metric..
        """
        qf = vtki.new("MeshQuality")
        qf.SetInputData(self.dataset)
        qf.SetTetQualityMeasure(metric)
        qf.SaveCellQualityOn()
        qf.Update()
        self._update(qf.GetOutput())
        return utils.vtk2numpy(qf.GetOutput().GetCellData().GetArray("Quality"))

    def check_validity(self, tol=0) -> np.ndarray:
        """
        Return an array of possible problematic tets following this convention:
        ```python
        Valid               =  0
        WrongNumberOfPoints = 01
        IntersectingEdges   = 02
        IntersectingFaces   = 04
        NoncontiguousEdges  = 08
        Nonconvex           = 10
        OrientedIncorrectly = 20
        ```

        Arguments:
            tol : (float)
                This value is used as an epsilon for floating point
                equality checks throughout the cell checking process.
        """
        vald = vtki.new("CellValidator")
        if tol:
            vald.SetTolerance(tol)
        vald.SetInputData(self.dataset)
        vald.Update()
        varr = vald.GetOutput().GetCellData().GetArray("ValidityState")
        return utils.vtk2numpy(varr)

    def decimate(self, scalars_name: str, fraction=0.5, n=0) -> Self:
        """
        Downsample the number of tets in a TetMesh to a specified fraction.
        Either `fraction` or `n` must be set.

        Arguments:
            fraction : (float)
                the desired final fraction of the total.
            n : (int)
                the desired number of final tets

        .. note:: setting `fraction=0.1` leaves 10% of the original nr of tets.
        """
        decimate = vtki.new("UnstructuredGridQuadricDecimation")
        decimate.SetInputData(self.dataset)
        decimate.SetScalarsName(scalars_name)

        if n:  # n = desired number of points
            decimate.SetNumberOfTetsOutput(n)
        else:
            decimate.SetTargetReduction(1 - fraction)
        decimate.Update()
        self._update(decimate.GetOutput())
        self.pipeline = utils.OperationNode(
            "decimate", comment=f"array: {scalars_name}", c="#edabab", parents=[self]
        )
        return self

    def subdivide(self) -> Self:
        """
        Increase the number of tetrahedrons of a `TetMesh`.
        Subdivides each tetrahedron into twelve smaller tetras.
        """
        sd = vtki.new("SubdivideTetra")
        sd.SetInputData(self.dataset)
        sd.Update()
        self._update(sd.GetOutput())
        self.pipeline = utils.OperationNode("subdivide", c="#edabab", parents=[self])
        return self

    def generate_random_points(self, n, min_radius=0) -> "vedo.Points":
        """
        Generate `n` uniformly distributed random points
        inside the tetrahedral mesh.

        A new point data array is added to the output points
        called "OriginalCellID" which contains the index of
        the cell ID in which the point was generated.

        Arguments:
            n : (int)
                number of points to generate.
            min_radius: (float)
                impose a minimum distance between points.
                If `min_radius` is set to 0, the points are
                generated uniformly at random inside the mesh.
                If `min_radius` is set to a positive value,
                the points are generated uniformly at random
                inside the mesh, but points closer than `min_radius`
                to any other point are discarded.

        Returns a `vedo.Points` object.

        Note:
            Consider using `points.probe(msh)` to interpolate
            any existing mesh data onto the points.

        Example:
        ```python
        from vedo import *
        tmesh = TetMesh(dataurl + "limb.vtu").alpha(0.2)
        pts = tmesh.generate_random_points(20000, min_radius=10)
        print(pts.pointdata["OriginalCellID"])
        show(pts, tmesh, axes=1).close()
        ```
        """
        cmesh = self.compute_cell_size()
        tets = cmesh.cells
        verts = cmesh.coordinates
        cumul = np.cumsum(np.abs(cmesh.celldata["Volume"]))

        out_pts = []
        orig_cell = []
        for _ in range(n):
            random_area = np.random.random() * cumul[-1]
            it = np.searchsorted(cumul, random_area)
            A, B, C, D = verts[tets[it]]
            r1, r2, r3 = sorted(np.random.random(3))
            p = r1 * A + (r2 - r1) * B + (r3 - r2) * C + (1 - r3) * D
            out_pts.append(p)
            orig_cell.append(it)
        orig_cellnp = np.array(orig_cell, dtype=np.uint32)

        vpts = vedo.pointcloud.Points(out_pts)
        vpts.pointdata["OriginalCellID"] = orig_cellnp

        if min_radius > 0:
            vpts.subsample(min_radius, absolute=True)

        vpts.point_size(5).color("k1")
        vpts.name = "RandomPoints"
        vpts.pipeline = utils.OperationNode(
            "generate_random_points", c="#edabab", parents=[self])
        return vpts

    def isosurface(self, value=True, flying_edges=None) -> "vedo.Mesh":
        """
        Return a `vedo.Mesh` isosurface.
        The "isosurface" is the surface of the region of points
        whose values equal to `value`.

        Set `value` to a single value or list of values to compute the isosurface(s).

        Note that flying_edges option is not available for `TetMesh`.
        """
        if flying_edges is not None:
            vedo.logger.warning("flying_edges option is not available for TetMesh.")

        if not self.dataset.GetPointData().GetScalars():
            vedo.logger.warning(
                "in isosurface() no scalar pointdata found. "
                "Mappping cells to points."
            )
            self.map_cells_to_points()
        scrange = self.dataset.GetPointData().GetScalars().GetRange()
        cf = vtki.new("ContourFilter")  # vtki.new("ContourGrid")
        cf.SetInputData(self.dataset)

        if utils.is_sequence(value):
            cf.SetNumberOfContours(len(value))
            for i, t in enumerate(value):
                cf.SetValue(i, t)
            cf.Update()
        else:
            if value is True:
                value = (2 * scrange[0] + scrange[1]) / 3.0
            cf.SetValue(0, value)
            cf.Update()

        msh = Mesh(cf.GetOutput(), c=None)
        msh.copy_properties_from(self)
        msh.pipeline = utils.OperationNode("isosurface", c="#edabab", parents=[self])
        return msh

    def slice(self, origin=(0, 0, 0), normal=(1, 0, 0)) -> "vedo.Mesh":
        """
        Return a 2D slice of the mesh by a plane passing through origin and assigned normal.
        """
        strn = str(normal)
        if   strn ==  "x": normal = (1, 0, 0)
        elif strn ==  "y": normal = (0, 1, 0)
        elif strn ==  "z": normal = (0, 0, 1)
        elif strn == "-x": normal = (-1, 0, 0)
        elif strn == "-y": normal = (0, -1, 0)
        elif strn == "-z": normal = (0, 0, -1)
        plane = vtki.new("Plane")
        plane.SetOrigin(origin)
        plane.SetNormal(normal)

        cc = vtki.new("Cutter")
        cc.SetInputData(self.dataset)
        cc.SetCutFunction(plane)
        cc.Update()
        msh = Mesh(cc.GetOutput()).flat().lighting("ambient")
        msh.copy_properties_from(self)
        msh.pipeline = utils.OperationNode("slice", c="#edabab", parents=[self])
        return msh


##########################################################################
class RectilinearGrid(PointAlgorithms, MeshVisual):
    """
    Build a rectilinear grid.
    """

    def __init__(self, inputobj=None):
        """
        A RectilinearGrid is a dataset where edges are parallel to the coordinate axes.
        It can be thought of as a tessellation of a box in 3D space, similar to a `Volume`
        except that the cells are not necessarily cubes, but they can have different lengths
        along each axis.
        This can be useful to describe a volume with variable resolution where one needs
        to represent a region with higher detail with respect to another region.

        Arguments:
            inputobj : (vtkRectilinearGrid, list, str)
                list of points and tet indices, or filename
        
        Example:
            ```python
            from vedo import RectilinearGrid, show

            xcoords = 7 + np.sqrt(np.arange(0,2500,25))
            ycoords = np.arange(0, 20)
            zcoords = np.arange(0, 20)

            rgrid = RectilinearGrid([xcoords, ycoords, zcoords])

            print(rgrid)
            print(rgrid.x_coordinates().shape)
            print(rgrid.compute_structured_coords([20,10,11]))

            msh = rgrid.tomesh().lw(1)

            show(msh, axes=1, viewup="z")
            ```
        """

        super().__init__()

        self.dataset = None

        self.mapper = vtki.new("PolyDataMapper")
        self._actor = vtki.vtkActor()
        self._actor.retrieve_object = weak_ref_to(self)
        self._actor.SetMapper(self.mapper)
        self.properties = self._actor.GetProperty()

        self.transform = LinearTransform()
        self.point_locator = None
        self.cell_locator = None
        self.line_locator = None

        self.name = "RectilinearGrid"
        self.filename = ""

        self.info = {}
        self.time =  time.time()

        ###############################
        if inputobj is None:
            self.dataset = vtki.vtkRectilinearGrid()

        elif isinstance(inputobj, vtki.vtkRectilinearGrid):
            self.dataset = inputobj

        elif isinstance(inputobj, RectilinearGrid):
            self.dataset = inputobj.dataset

        elif isinstance(inputobj, str):
            if "https://" in inputobj:
                inputobj = download(inputobj, verbose=False)
            if inputobj.endswith(".vtr"):
                reader = vtki.new("XMLRectilinearGridReader")
            else:
                reader = vtki.new("RectilinearGridReader")
            self.filename = inputobj
            reader.SetFileName(inputobj)
            reader.Update()
            self.dataset = reader.GetOutput()
        
        elif utils.is_sequence(inputobj):
            self.dataset = vtki.vtkRectilinearGrid()
            xcoords, ycoords, zcoords = inputobj
            nx, ny, nz = len(xcoords), len(ycoords), len(zcoords)
            self.dataset.SetDimensions(nx, ny, nz)
            self.dataset.SetXCoordinates(utils.numpy2vtk(xcoords))
            self.dataset.SetYCoordinates(utils.numpy2vtk(ycoords))
            self.dataset.SetZCoordinates(utils.numpy2vtk(zcoords))

        ###############################

        if not self.dataset:
            vedo.logger.error(f"RectilinearGrid: cannot understand input type {type(inputobj)}")
            return

        self.properties.SetColor(0.352, 0.612, 0.996)  # blue7

        self.pipeline = utils.OperationNode(
            self, comment=f"#tets {self.dataset.GetNumberOfCells()}", c="#9e2a2b"
        )

    @property
    def actor(self):
        """Return the `vtkActor` of the object."""
        gf = vtki.new("GeometryFilter")
        gf.SetInputData(self.dataset)
        gf.Update()
        self.mapper.SetInputData(gf.GetOutput())
        self.mapper.Modified()
        return self._actor

    @actor.setter
    def actor(self, _):
        pass

    def _update(self, data, reset_locators=False):
        self.dataset = data
        if reset_locators:
            self.cell_locator = None
            self.point_locator = None
        return self

    ##################################################################
    def __str__(self):
        """Print a summary for the `RectilinearGrid` object."""
        module = self.__class__.__module__
        name = self.__class__.__name__
        out = vedo.printc(
            f"{module}.{name} at ({hex(self.memory_address())})".ljust(75),
            c="c", bold=True, invert=True, return_string=True,
        )
        out += "\x1b[0m\x1b[36;1m"

        out += "name".ljust(14) + ": " + str(self.name) + "\n"
        if self.filename:
            out += "filename".ljust(14) + ": " + str(self.filename) + "\n"

        out += "dimensions".ljust(14) + ": " + str(self.dataset.GetDimensions()) + "\n"

        out += "center".ljust(14) + ": "
        out += utils.precision(self.dataset.GetCenter(), 6) + "\n"

        bnds = self.bounds()
        bx1, bx2 = utils.precision(bnds[0], 3), utils.precision(bnds[1], 3)
        by1, by2 = utils.precision(bnds[2], 3), utils.precision(bnds[3], 3)
        bz1, bz2 = utils.precision(bnds[4], 3), utils.precision(bnds[5], 3)
        out += "bounds".ljust(14) + ":"
        out += " x=(" + bx1 + ", " + bx2 + "),"
        out += " y=(" + by1 + ", " + by2 + "),"
        out += " z=(" + bz1 + ", " + bz2 + ")\n"

        out += "memory size".ljust(14) + ": "
        out += utils.precision(self.dataset.GetActualMemorySize() / 1024, 2) + " MB\n"

        for key in self.pointdata.keys():
            arr = self.pointdata[key]
            rng = utils.precision(arr.min(), 3) + ", " + utils.precision(arr.max(), 3)
            mark_active = "pointdata"
            a_scalars = self.dataset.GetPointData().GetScalars()
            a_vectors = self.dataset.GetPointData().GetVectors()
            a_tensors = self.dataset.GetPointData().GetTensors()
            if a_scalars and a_scalars.GetName() == key:
                mark_active += " *"
            elif a_vectors and a_vectors.GetName() == key:
                mark_active += " **"
            elif a_tensors and a_tensors.GetName() == key:
                mark_active += " ***"
            out += mark_active.ljust(14) + f': "{key}" ({arr.dtype}), ndim={arr.ndim}'
            out += f", range=({rng})\n"

        for key in self.celldata.keys():
            arr = self.celldata[key]
            rng = utils.precision(arr.min(), 3) + ", " + utils.precision(arr.max(), 3)
            mark_active = "celldata"
            a_scalars = self.dataset.GetCellData().GetScalars()
            a_vectors = self.dataset.GetCellData().GetVectors()
            a_tensors = self.dataset.GetCellData().GetTensors()
            if a_scalars and a_scalars.GetName() == key:
                mark_active += " *"
            elif a_vectors and a_vectors.GetName() == key:
                mark_active += " **"
            elif a_tensors and a_tensors.GetName() == key:
                mark_active += " ***"
            out += mark_active.ljust(14) + f': "{key}" ({arr.dtype}), ndim={arr.ndim}'
            out += f", range=({rng})\n"

        for key in self.metadata.keys():
            arr = self.metadata[key]
            out += "metadata".ljust(14) + ": " + f'"{key}" ({len(arr)} values)\n'

        return out.rstrip() + "\x1b[0m"

    def _repr_html_(self):
        """
        HTML representation of the RectilinearGrid object for Jupyter Notebooks.

        Returns:
            HTML text with the image and some properties.
        """
        import io
        import base64
        from PIL import Image

        library_name = "vedo.grids.RectilinearGrid"
        help_url = "https://vedo.embl.es/docs/vedo/grids.html#RectilinearGrid"

        m = self.tomesh().linewidth(1).lighting("off")
        arr= m.thumbnail(zoom=1, elevation=-30, azimuth=-30)

        im = Image.fromarray(arr)
        buffered = io.BytesIO()
        im.save(buffered, format="PNG", quality=100)
        encoded = base64.b64encode(buffered.getvalue()).decode("utf-8")
        url = "data:image/png;base64," + encoded
        image = f"<img src='{url}'></img>"

        bounds = "<br/>".join(
            [
                utils.precision(min_x, 4) + " ... " + utils.precision(max_x, 4)
                for min_x, max_x in zip(self.bounds()[::2], self.bounds()[1::2])
            ]
        )

        help_text = ""
        if self.name:
            help_text += f"<b> {self.name}: &nbsp&nbsp</b>"
        help_text += '<b><a href="' + help_url + '" target="_blank">' + library_name + "</a></b>"
        if self.filename:
            dots = ""
            if len(self.filename) > 30:
                dots = "..."
            help_text += f"<br/><code><i>({dots}{self.filename[-30:]})</i></code>"

        pdata = ""
        if self.dataset.GetPointData().GetScalars():
            if self.dataset.GetPointData().GetScalars().GetName():
                name = self.dataset.GetPointData().GetScalars().GetName()
                pdata = "<tr><td><b> point data array </b></td><td>" + name + "</td></tr>"

        cdata = ""
        if self.dataset.GetCellData().GetScalars():
            if self.dataset.GetCellData().GetScalars().GetName():
                name = self.dataset.GetCellData().GetScalars().GetName()
                cdata = "<tr><td><b> cell data array </b></td><td>" + name + "</td></tr>"

        pts = self.coordinates
        cm = np.mean(pts, axis=0)

        all = [
            "<table>",
            "<tr>",
            "<td>", image, "</td>",
            "<td style='text-align: center; vertical-align: center;'><br/>", help_text,
            "<table>",
            "<tr><td><b> bounds </b> <br/> (x/y/z) </td><td>" + str(bounds) + "</td></tr>",
            "<tr><td><b> center of mass </b></td><td>" + utils.precision(cm,3) + "</td></tr>",
            "<tr><td><b> nr. points&nbsp/&nbspcells </b></td><td>"
            + str(self.npoints) + "&nbsp/&nbsp" + str(self.ncells) + "</td></tr>",
            pdata,
            cdata,
            "</table>",
            "</table>",
        ]
        return "\n".join(all)

    def dimensions(self) -> np.ndarray:
        """Return the number of points in the x, y and z directions."""
        return np.array(self.dataset.GetDimensions())

    def x_coordinates(self) -> np.ndarray:
        """Return the x-coordinates of the grid."""
        return utils.vtk2numpy(self.dataset.GetXCoordinates())
    
    def y_coordinates(self) -> np.ndarray:
        """Return the y-coordinates of the grid."""
        return utils.vtk2numpy(self.dataset.GetYCoordinates())
    
    def z_coordinates(self) -> np.ndarray:
        """Return the z-coordinates of the grid."""
        return utils.vtk2numpy(self.dataset.GetZCoordinates())
    
    def is_point_visible(self, pid: int) -> bool:
        """Return True if point `pid` is visible."""
        return self.dataset.IsPointVisible(pid)
    
    def is_cell_visible(self, cid: int) -> bool:
        """Return True if cell `cid` is visible."""
        return self.dataset.IsCellVisible(cid)
    
    def has_blank_points(self) -> bool:
        """Return True if the grid has blank points."""
        return self.dataset.HasAnyBlankPoints()
    
    def has_blank_cells(self) -> bool:
        """Return True if the grid has blank cells."""
        return self.dataset.HasAnyBlankCells()
    
    def compute_structured_coords(self, x: list) -> dict:
        """
        Convenience function computes the structured coordinates for a point `x`.

        This method returns a dictionary with keys `ijk`, `pcoords` and `inside`.
        The cell is specified by the array `ijk`.
        and the parametric coordinates in the cell are specified with `pcoords`. 
        Value of `inside` is False if the point x is outside of the grid.
        """
        ijk = [0, 0, 0]
        pcoords = [0., 0., 0.]
        inout = self.dataset.ComputeStructuredCoordinates(x, ijk, pcoords)
        return {"ijk": np.array(ijk), "pcoords": np.array(pcoords), "inside": bool(inout)}
    
    def compute_pointid(self, ijk: int) -> int:
        """Given a location in structured coordinates (i-j-k), return the point id."""
        return self.dataset.ComputePointId(ijk)
    
    def compute_cellid(self, ijk: int) -> int:
        """Given a location in structured coordinates (i-j-k), return the cell id."""
        return self.dataset.ComputeCellId(ijk)
    
    def find_point(self, x: list) -> int:
        """Given a position `x`, return the id of the closest point."""
        return self.dataset.FindPoint(x)
    
    def find_cell(self, x: list) -> dict:
        """Given a position `x`, return the id of the closest cell."""
        cell = vtki.vtkHexagonalPrism()
        cellid = vtki.mutable(0)
        tol2 = 0.001 # vtki.mutable(0)
        subid = vtki.mutable(0)
        pcoords = [0.0, 0.0, 0.0]
        weights = [0.0, 0.0, 0.0]
        res = self.dataset.FindCell(x, cell, cellid, tol2, subid, pcoords, weights)
        result = {}
        result["cellid"] = cellid
        result["subid"] = subid
        result["pcoords"] = pcoords
        result["weights"] = weights
        result["status"] = res
        return result

    def clone(self, deep=True) -> "RectilinearGrid":
        """Return a clone copy of the RectilinearGrid. Alias of `copy()`."""
        if deep:
            newrg = vtki.vtkRectilinearGrid()
            newrg.CopyStructure(self.dataset)
            newrg.CopyAttributes(self.dataset)
            newvol = RectilinearGrid(newrg)
        else:
            newvol = RectilinearGrid(self.dataset)

        prop = vtki.vtkProperty()
        prop.DeepCopy(self.properties)
        newvol.actor.SetProperty(prop)
        newvol.properties = prop
        newvol.pipeline = utils.OperationNode("clone", parents=[self], c="#bbd0ff", shape="diamond")
        return newvol

    def bounds(self) -> np.ndarray:
        """
        Get the object bounds.
        Returns a list in format `[xmin,xmax, ymin,ymax, zmin,zmax]`.
        """
        # OVERRIDE CommonAlgorithms.bounds() which is too slow
        return np.array(self.dataset.GetBounds())

    def isosurface(self, value=None) -> "vedo.Mesh":
        """
        Return a `Mesh` isosurface extracted from the object.

        Set `value` as single float or list of values to draw the isosurface(s).
        """
        scrange = self.dataset.GetScalarRange()

        cf = vtki.new("ContourFilter")
        cf.UseScalarTreeOn()
        cf.SetInputData(self.dataset)
        cf.ComputeNormalsOn()

        if value is None:
            value = (2 * scrange[0] + scrange[1]) / 3.0
            # print("automatic isosurface value =", value)
            cf.SetValue(0, value)
        else:
            if utils.is_sequence(value):
                cf.SetNumberOfContours(len(value))
                for i, t in enumerate(value):
                    cf.SetValue(i, t)
            else:
                cf.SetValue(0, value)

        cf.Update()
        poly = cf.GetOutput()

        out = vedo.mesh.Mesh(poly, c=None).phong()
        out.mapper.SetScalarRange(scrange[0], scrange[1])

        out.pipeline = utils.OperationNode(
            "isosurface",
            parents=[self],
            comment=f"#pts {out.dataset.GetNumberOfPoints()}",
            c="#4cc9f0:#e9c46a",
        )
        return out

    def cut_with_plane(self, origin=(0, 0, 0), normal="x") -> "UnstructuredGrid":
        """
        Cut the object with the plane defined by a point and a normal.

        Arguments:
            origin : (list)
                the cutting plane goes through this point
            normal : (list, str)
                normal vector to the cutting plane
        """
        strn = str(normal)
        if strn   ==  "x": normal = (1, 0, 0)
        elif strn ==  "y": normal = (0, 1, 0)
        elif strn ==  "z": normal = (0, 0, 1)
        elif strn == "-x": normal = (-1, 0, 0)
        elif strn == "-y": normal = (0, -1, 0)
        elif strn == "-z": normal = (0, 0, -1)
        plane = vtki.new("Plane")
        plane.SetOrigin(origin)
        plane.SetNormal(normal)
        clipper = vtki.new("ClipDataSet")
        clipper.SetInputData(self.dataset)
        clipper.SetClipFunction(plane)
        clipper.GenerateClipScalarsOff()
        clipper.GenerateClippedOutputOff()
        clipper.SetValue(0)
        clipper.Update()
        cout = clipper.GetOutput()
        ug = vedo.UnstructuredGrid(cout)
        if isinstance(self, UnstructuredGrid):
            self._update(cout)
            self.pipeline = utils.OperationNode("cut_with_plane", parents=[self], c="#9e2a2b")
            return self
        ug.pipeline = utils.OperationNode("cut_with_plane", parents=[self], c="#9e2a2b")
        return ug

    def cut_with_mesh(self, mesh, invert=False, whole_cells=False, on_boundary=False) -> "UnstructuredGrid":
        """
        Cut a `RectilinearGrid` with a `Mesh`.

        Use `invert` to return cut off part of the input object.
        """
        ug = self.dataset

        ippd = vtki.new("ImplicitPolyDataDistance")
        ippd.SetInput(mesh.dataset)

        if whole_cells or on_boundary:
            clipper = vtki.new("ExtractGeometry")
            clipper.SetInputData(ug)
            clipper.SetImplicitFunction(ippd)
            clipper.SetExtractInside(not invert)
            clipper.SetExtractBoundaryCells(False)
            if on_boundary:
                clipper.SetExtractBoundaryCells(True)
                clipper.SetExtractOnlyBoundaryCells(True)
        else:
            signed_dists = vtki.vtkFloatArray()
            signed_dists.SetNumberOfComponents(1)
            signed_dists.SetName("SignedDistance")
            for pointId in range(ug.GetNumberOfPoints()):
                p = ug.GetPoint(pointId)
                signed_dist = ippd.EvaluateFunction(p)
                signed_dists.InsertNextValue(signed_dist)
            ug.GetPointData().AddArray(signed_dists)
            ug.GetPointData().SetActiveScalars("SignedDistance")  # NEEDED
            clipper = vtki.new("ClipDataSet")
            clipper.SetInputData(ug)
            clipper.SetInsideOut(not invert)
            clipper.SetValue(0.0)

        clipper.Update()

        out = UnstructuredGrid(clipper.GetOutput())
        out.pipeline = utils.OperationNode("cut_with_mesh", parents=[self], c="#9e2a2b")
        return out


##########################################################################
class StructuredGrid(PointAlgorithms, MeshVisual):
    """
    Build a structured grid.
    """

    def __init__(self, inputobj=None):
        """
        A StructuredGrid is a dataset where edges of the hexahedrons are 
        not necessarily parallel to the coordinate axes.
        It can be thought of as a tessellation of a block of 3D space,
        similar to a `RectilinearGrid`
        except that the cells are not necessarily cubes, they can have different
        orientations but are connected in the same way as a `RectilinearGrid`.

        Arguments:
            inputobj : (vtkStructuredGrid, list, str)
                list of points and tet indices, or filename
        
        Example:
            ```python
            from vedo import *
            sgrid = StructuredGrid(dataurl+"structgrid.vts")
            print(sgrid)
            msh = sgrid.tomesh().lw(1)
            show(msh, axes=1, viewup="z")
            ```

            ```python
            from vedo import *

            cx = np.sqrt(np.linspace(100, 400, 10))
            cy = np.linspace(30, 40, 20)
            cz = np.linspace(40, 50, 30)
            x, y, z = np.meshgrid(cx, cy, cz)

            sgrid1 = StructuredGrid([x, y, z])
            sgrid1.cmap("viridis", sgrid1.coordinates[:, 0])
            print(sgrid1)

            sgrid2 = sgrid1.clone().cut_with_plane(normal=(-1,1,1), origin=[14,34,44])
            msh2 = sgrid2.tomesh(shrink=0.9).lw(1).cmap("viridis")

            show(
                [["StructuredGrid", sgrid1], ["Shrinked Mesh", msh2]],
                N=2, axes=1, viewup="z",
            )
            ```
        """

        super().__init__()

        self.dataset = None

        self.mapper = vtki.new("PolyDataMapper")
        self._actor = vtki.vtkActor()
        self._actor.retrieve_object = weak_ref_to(self)
        self._actor.SetMapper(self.mapper)
        self.properties = self._actor.GetProperty()

        self.transform = LinearTransform()
        self.point_locator = None
        self.cell_locator = None
        self.line_locator = None

        self.name = "StructuredGrid"
        self.filename = ""

        self.info = {}
        self.time =  time.time()

        ###############################
        if inputobj is None:
            self.dataset = vtki.vtkStructuredGrid()

        elif isinstance(inputobj, vtki.vtkStructuredGrid):
            self.dataset = inputobj

        elif isinstance(inputobj, StructuredGrid):
            self.dataset = inputobj.dataset

        elif isinstance(inputobj, str):
            if "https://" in inputobj:
                inputobj = download(inputobj, verbose=False)
            if inputobj.endswith(".vts"):
                reader = vtki.new("XMLStructuredGridReader")
            else:
                reader = vtki.new("StructuredGridReader")
            self.filename = inputobj
            reader.SetFileName(inputobj)
            reader.Update()
            self.dataset = reader.GetOutput()
        
        elif utils.is_sequence(inputobj):
            self.dataset = vtki.vtkStructuredGrid()
            x, y, z = inputobj
            xyz = np.vstack((
                x.flatten(order="F"),
                y.flatten(order="F"),
                z.flatten(order="F"))
            ).T
            dims = x.shape
            self.dataset.SetDimensions(dims)      
            # self.dataset.SetDimensions(dims[1], dims[0], dims[2])
            vpoints = vtki.vtkPoints()
            vpoints.SetData(utils.numpy2vtk(xyz))
            self.dataset.SetPoints(vpoints)


        ###############################
        if not self.dataset:
            vedo.logger.error(f"StructuredGrid: cannot understand input type {type(inputobj)}")
            return

        self.properties.SetColor(0.352, 0.612, 0.996)  # blue7

        self.pipeline = utils.OperationNode(
            self, comment=f"#tets {self.dataset.GetNumberOfCells()}", c="#9e2a2b"
        )

    @property
    def actor(self):
        """Return the `vtkActor` of the object."""
        gf = vtki.new("GeometryFilter")
        gf.SetInputData(self.dataset)
        gf.Update()
        self.mapper.SetInputData(gf.GetOutput())
        self.mapper.Modified()
        return self._actor

    @actor.setter
    def actor(self, _):
        pass

    def _update(self, data, reset_locators=False):
        self.dataset = data
        if reset_locators:
            self.cell_locator = None
            self.point_locator = None
        return self

    ##################################################################
    def __str__(self):
        """Print a summary for the `StructuredGrid` object."""
        module = self.__class__.__module__
        name = self.__class__.__name__
        out = vedo.printc(
            f"{module}.{name} at ({hex(self.memory_address())})".ljust(75),
            c="c", bold=True, invert=True, return_string=True,
        )
        out += "\x1b[0m\x1b[36;1m"

        out += "name".ljust(14) + ": " + str(self.name) + "\n"
        if self.filename:
            out += "filename".ljust(14) + ": " + str(self.filename) + "\n"

        out += "dimensions".ljust(14) + ": " + str(self.dataset.GetDimensions()) + "\n"

        out += "center".ljust(14) + ": "
        out += utils.precision(self.dataset.GetCenter(), 6) + "\n"

        bnds = self.bounds()
        bx1, bx2 = utils.precision(bnds[0], 3), utils.precision(bnds[1], 3)
        by1, by2 = utils.precision(bnds[2], 3), utils.precision(bnds[3], 3)
        bz1, bz2 = utils.precision(bnds[4], 3), utils.precision(bnds[5], 3)
        out += "bounds".ljust(14) + ":"
        out += " x=(" + bx1 + ", " + bx2 + "),"
        out += " y=(" + by1 + ", " + by2 + "),"
        out += " z=(" + bz1 + ", " + bz2 + ")\n"

        out += "memory size".ljust(14) + ": "
        out += utils.precision(self.dataset.GetActualMemorySize() / 1024, 2) + " MB\n"

        for key in self.pointdata.keys():
            arr = self.pointdata[key]
            rng = utils.precision(arr.min(), 3) + ", " + utils.precision(arr.max(), 3)
            mark_active = "pointdata"
            a_scalars = self.dataset.GetPointData().GetScalars()
            a_vectors = self.dataset.GetPointData().GetVectors()
            a_tensors = self.dataset.GetPointData().GetTensors()
            if a_scalars and a_scalars.GetName() == key:
                mark_active += " *"
            elif a_vectors and a_vectors.GetName() == key:
                mark_active += " **"
            elif a_tensors and a_tensors.GetName() == key:
                mark_active += " ***"
            out += mark_active.ljust(14) + f': "{key}" ({arr.dtype}), ndim={arr.ndim}'
            out += f", range=({rng})\n"

        for key in self.celldata.keys():
            arr = self.celldata[key]
            rng = utils.precision(arr.min(), 3) + ", " + utils.precision(arr.max(), 3)
            mark_active = "celldata"
            a_scalars = self.dataset.GetCellData().GetScalars()
            a_vectors = self.dataset.GetCellData().GetVectors()
            a_tensors = self.dataset.GetCellData().GetTensors()
            if a_scalars and a_scalars.GetName() == key:
                mark_active += " *"
            elif a_vectors and a_vectors.GetName() == key:
                mark_active += " **"
            elif a_tensors and a_tensors.GetName() == key:
                mark_active += " ***"
            out += mark_active.ljust(14) + f': "{key}" ({arr.dtype}), ndim={arr.ndim}'
            out += f", range=({rng})\n"

        for key in self.metadata.keys():
            arr = self.metadata[key]
            out += "metadata".ljust(14) + ": " + f'"{key}" ({len(arr)} values)\n'

        return out.rstrip() + "\x1b[0m"
    
    def _repr_html_(self):
        """
        HTML representation of the StructuredGrid object for Jupyter Notebooks.

        Returns:
            HTML text with the image and some properties.
        """
        import io
        import base64
        from PIL import Image

        library_name = "vedo.grids.StructuredGrid"
        help_url = "https://vedo.embl.es/docs/vedo/grids.html#StructuredGrid"

        m = self.tomesh().linewidth(1).lighting("off")
        arr= m.thumbnail(zoom=1, elevation=-30, azimuth=-30)

        im = Image.fromarray(arr)
        buffered = io.BytesIO()
        im.save(buffered, format="PNG", quality=100)
        encoded = base64.b64encode(buffered.getvalue()).decode("utf-8")
        url = "data:image/png;base64," + encoded
        image = f"<img src='{url}'></img>"

        bounds = "<br/>".join(
            [
                utils.precision(min_x, 4) + " ... " + utils.precision(max_x, 4)
                for min_x, max_x in zip(self.bounds()[::2], self.bounds()[1::2])
            ]
        )

        help_text = ""
        if self.name:
            help_text += f"<b> {self.name}: &nbsp&nbsp</b>"
        help_text += '<b><a href="' + help_url + '" target="_blank">' + library_name + "</a></b>"
        if self.filename:
            dots = ""
            if len(self.filename) > 30:
                dots = "..."
            help_text += f"<br/><code><i>({dots}{self.filename[-30:]})</i></code>"

        pdata = ""
        if self.dataset.GetPointData().GetScalars():
            if self.dataset.GetPointData().GetScalars().GetName():
                name = self.dataset.GetPointData().GetScalars().GetName()
                pdata = "<tr><td><b> point data array </b></td><td>" + name + "</td></tr>"

        cdata = ""
        if self.dataset.GetCellData().GetScalars():
            if self.dataset.GetCellData().GetScalars().GetName():
                name = self.dataset.GetCellData().GetScalars().GetName()
                cdata = "<tr><td><b> cell data array </b></td><td>" + name + "</td></tr>"

        pts = self.coordinates
        cm = np.mean(pts, axis=0)

        all = [
            "<table>",
            "<tr>",
            "<td>", image, "</td>",
            "<td style='text-align: center; vertical-align: center;'><br/>", help_text,
            "<table>",
            "<tr><td><b> bounds </b> <br/> (x/y/z) </td><td>" + str(bounds) + "</td></tr>",
            "<tr><td><b> center of mass </b></td><td>" + utils.precision(cm,3) + "</td></tr>",
            "<tr><td><b> nr. points&nbsp/&nbspcells </b></td><td>"
            + str(self.npoints) + "&nbsp/&nbsp" + str(self.ncells) + "</td></tr>",
            pdata,
            cdata,
            "</table>",
            "</table>",
        ]
        return "\n".join(all)

    def dimensions(self) -> np.ndarray:
        """Return the number of points in the x, y and z directions."""
        return np.array(self.dataset.GetDimensions())

    def clone(self, deep=True) -> "StructuredGrid":
        """Return a clone copy of the StructuredGrid. Alias of `copy()`."""
        if deep:
            newrg = vtki.vtkStructuredGrid()
            newrg.CopyStructure(self.dataset)
            newrg.CopyAttributes(self.dataset)
            newvol = StructuredGrid(newrg)
        else:
            newvol = StructuredGrid(self.dataset)

        prop = vtki.vtkProperty()
        prop.DeepCopy(self.properties)
        newvol.actor.SetProperty(prop)
        newvol.properties = prop
        newvol.pipeline = utils.OperationNode("clone", parents=[self], c="#bbd0ff", shape="diamond")
        return newvol

    def find_point(self, x: list) -> int:
        """Given a position `x`, return the id of the closest point."""
        return self.dataset.FindPoint(x)
    
    def find_cell(self, x: list) -> dict:
        """Given a position `x`, return the id of the closest cell."""
        cell = vtki.vtkHexagonalPrism()
        cellid = vtki.mutable(0)
        tol2 = 0.001 # vtki.mutable(0)
        subid = vtki.mutable(0)
        pcoords = [0.0, 0.0, 0.0]
        weights = [0.0, 0.0, 0.0]
        res = self.dataset.FindCell(x, cell, cellid, tol2, subid, pcoords, weights)
        result = {}
        result["cellid"] = cellid
        result["subid"] = subid
        result["pcoords"] = pcoords
        result["weights"] = weights
        result["status"] = res
        return result

    def cut_with_plane(self, origin=(0, 0, 0), normal="x") -> "vedo.UnstructuredGrid":
        """
        Cut the object with the plane defined by a point and a normal.

        Arguments:
            origin : (list)
                the cutting plane goes through this point
            normal : (list, str)
                normal vector to the cutting plane
        
        Returns an `UnstructuredGrid` object.
        """
        strn = str(normal)
        if strn   ==  "x": normal = (1, 0, 0)
        elif strn ==  "y": normal = (0, 1, 0)
        elif strn ==  "z": normal = (0, 0, 1)
        elif strn == "-x": normal = (-1, 0, 0)
        elif strn == "-y": normal = (0, -1, 0)
        elif strn == "-z": normal = (0, 0, -1)
        plane = vtki.new("Plane")
        plane.SetOrigin(origin)
        plane.SetNormal(normal)
        clipper = vtki.new("ClipDataSet")
        clipper.SetInputData(self.dataset)
        clipper.SetClipFunction(plane)
        clipper.GenerateClipScalarsOff()
        clipper.GenerateClippedOutputOff()
        clipper.SetValue(0)
        clipper.Update()
        cout = clipper.GetOutput()
        ug = vedo.UnstructuredGrid(cout)
        if isinstance(self, vedo.UnstructuredGrid):
            self._update(cout)
            self.pipeline = utils.OperationNode("cut_with_plane", parents=[self], c="#9e2a2b")
            return self
        ug.pipeline = utils.OperationNode("cut_with_plane", parents=[self], c="#9e2a2b")
        return ug

    def cut_with_mesh(self, mesh: Mesh, invert=False, whole_cells=False, on_boundary=False) -> "UnstructuredGrid":
        """
        Cut a `RectilinearGrid` with a `Mesh`.

        Use `invert` to return cut off part of the input object.

        Returns an `UnstructuredGrid` object.
        """
        ug = self.dataset

        ippd = vtki.new("ImplicitPolyDataDistance")
        ippd.SetInput(mesh.dataset)

        if whole_cells or on_boundary:
            clipper = vtki.new("ExtractGeometry")
            clipper.SetInputData(ug)
            clipper.SetImplicitFunction(ippd)
            clipper.SetExtractInside(not invert)
            clipper.SetExtractBoundaryCells(False)
            if on_boundary:
                clipper.SetExtractBoundaryCells(True)
                clipper.SetExtractOnlyBoundaryCells(True)
        else:
            signed_dists = vtki.vtkFloatArray()
            signed_dists.SetNumberOfComponents(1)
            signed_dists.SetName("SignedDistance")
            for pointId in range(ug.GetNumberOfPoints()):
                p = ug.GetPoint(pointId)
                signed_dist = ippd.EvaluateFunction(p)
                signed_dists.InsertNextValue(signed_dist)
            ug.GetPointData().AddArray(signed_dists)
            ug.GetPointData().SetActiveScalars("SignedDistance")  # NEEDED
            clipper = vtki.new("ClipDataSet")
            clipper.SetInputData(ug)
            clipper.SetInsideOut(not invert)
            clipper.SetValue(0.0)

        clipper.Update()

        out = UnstructuredGrid(clipper.GetOutput())
        out.pipeline = utils.OperationNode("cut_with_mesh", parents=[self], c="#9e2a2b")
        return out

    def isosurface(self, value=None) -> "vedo.Mesh":
        """
        Return a `Mesh` isosurface extracted from the object.

        Set `value` as single float or list of values to draw the isosurface(s).
        """
        scrange = self.dataset.GetScalarRange()

        cf = vtki.new("ContourFilter")
        cf.UseScalarTreeOn()
        cf.SetInputData(self.dataset)
        cf.ComputeNormalsOn()

        if value is None:
            value = (2 * scrange[0] + scrange[1]) / 3.0
            # print("automatic isosurface value =", value)
            cf.SetValue(0, value)
        else:
            if utils.is_sequence(value):
                cf.SetNumberOfContours(len(value))
                for i, t in enumerate(value):
                    cf.SetValue(i, t)
            else:
                cf.SetValue(0, value)

        cf.Update()
        poly = cf.GetOutput()

        out = vedo.mesh.Mesh(poly, c=None).phong()
        out.mapper.SetScalarRange(scrange[0], scrange[1])

        out.pipeline = utils.OperationNode(
            "isosurface",
            parents=[self],
            comment=f"#pts {out.dataset.GetNumberOfPoints()}",
            c="#4cc9f0:#e9c46a",
        )
        return out



from vedo import *


def render_slice(vslice, name):
    vslice.cut_with_scalar(rmin, "input_scalars", invert=True)
    vslice.triangulate()
    vslice.cmap(cmap_slicer, vmin=rmin, vmax=rmax).lighting("off")
    isos = vslice.isolines(vmin=rmin, vmax=rmax, n=12).c("black")
    vslice.name = name
    isos.name = name
    plt.remove(name).add(vslice, isos)

def slider_function_x(widget, event):
    i = int(widget.value)
    if i == widget.previous_value:
        return
    widget.previous_value = i
    render_slice(vol.xslice(i), "XSlice")

def slider_function_y(widget, event):
    j = int(widget.value)
    if j == widget.previous_value:
        return
    widget.previous_value = j
    render_slice(vol.yslice(j), "YSlice")


def slider_function_z(widget, event):
    k = int(widget.value)
    if k == widget.previous_value:
        return
    widget.previous_value = k
    render_slice(vol.zslice(k), "ZSlice")


if __name__ == "__main__":

    settings.default_font = "Roboto"
    cmap_slicer = "RdBu"

    datapath = download(dataurl+"geo_dataset.npy")
    dataset = np.load(datapath)
    min_value = np.nanmin(dataset)
    max_value = np.nanmax(dataset)

    rmin = np.nanquantile(dataset, q=0.30)
    rmax = np.nanquantile(dataset, q=0.95)

    # replace NaNs with a value to mask them in the rendered window
    nan_ind = np.isnan(dataset)
    dataset[nan_ind] = 0

    vol = Volume(dataset, spacing=[15, 15, 2])
    dims = vol.dimensions()

    iso = vol.isosurface(rmin).smooth()
    iso.cmap(cmap_slicer, vmin=min_value, vmax=max_value)
    iso.add_scalarbar3d(c="black", title="scalar value")
    iso.scalarbar = iso.scalarbar.clone2d("center-right", size=0.2)
    iso.c("k5").alpha(0.1).lighting("off").wireframe().pickable(False).backface_culling()

    plt = Plotter(size=(1400, 1200))

    plt.add_slider(
        slider_function_x,
        0, dims[0],
        pos=[(0.7, 0.12), (0.95, 0.12)],
        show_value=False,
        c="dr",
    )

    plt.add_slider(
        slider_function_y,
        0, dims[1],
        pos=[(0.7, 0.08), (0.95, 0.08)],
        show_value=False,
        c="dg",
    )

    plt.add_slider(
        slider_function_z,
        0, dims[2],
        pos=[(0.7, 0.04), (0.95, 0.04)],
        show_value=False,
        c="db",
    )

    plt.show(iso, viewup="z", axes=1).close()



# Same as warp4b.py but using the applications.MorphPlotter class
from vedo import Mesh, settings, dataurl
from vedo.applications import MorphPlotter


####################################################################################
# THIS IS IMPLEMENTED IN vedo.applications.MorphPlotter, shown here for reference
####################################################################################
# from vedo import Plotter, Points, Text2D, Axes
# class MorphPlotter(Plotter):
#     
#     def __init__(self, source, target, **kwargs):
#         kwargs.update(dict(N=3, sharecam=0))
#         super().__init__(**kwargs)
#
#         self.source = source.pickable(True)
#         self.target = target.pickable(False)
#         self.clicked = []
#         self.sources = []
#         self.targets = []
#         self.msg0 = Text2D("Pick a point on the surface",
#                            pos="bottom-center", c='white', bg="blue4", alpha=1)
#         self.msg1 = Text2D(pos="bottom-center", c='white', bg="blue4", alpha=1)
#         instructions = (
#             "Morphological alignment of 3D surfaces.\n"
#             "Pick a point on the source surface, then\n"
#             "pick the corresponding point on the target surface\n"
#             "Pick at least 4 point pairs. Press:\n"
#             "- c to clear the selection.\n"
#             "- d to delete the last selection.\n"
#             "- q to quit."
#         )
#         self.instructions = Text2D(instructions, s=0.7, bg="blue4", alpha=0.1)
#         self.at(0).add(source, self.msg0, self.instructions).reset_camera()
#         self.at(1).add(f"Reference {target.filename}", self.msg1, target)
#         cam1 = self.camera  # save camera at 1
#         self.at(2).add("Morphing Output", target).background("k9")
#         self.camera = cam1  # use the same camera of renderer1
#
#         self.callid1 = self.add_callback("on key press", self.on_keypress)
#         self.callid2 = self.add_callback("on click", self.on_click)
#         self._interactive = True
#
#     def update(self):
#         source_pts = Points(self.sources).color("purple5").ps(12)
#         target_pts = Points(self.targets).color("purple5").ps(12)
#         source_pts.name = "source_pts"
#         target_pts.name = "target_pts"
#         slabels = self.source_pts.labels2d("id", c="purple3")
#         tlabels = self.target_pts.labels2d("id", c="purple3")
#         slabels.name = "source_pts"
#         tlabels.name = "target_pts"
#         self.at(0).remove("source_pts").add(source_pts, slabels)
#         self.at(1).remove("target_pts").add(target_pts, tlabels)
#         self.render()
#
#         if len(self.sources) == len(self.targets) and len(self.sources) > 3:
#             warped = self.source.clone().warp(self.sources, self.targets)
#             warped.name = "warped"
#             self.at(2).remove("warped").add(warped)
#             self.render()
#
#     def on_click(self, evt):
#         if evt.object == source:
#             self.sources.append(evt.picked3d)
#             self.source.pickable(False)
#             self.target.pickable(True)
#             self.msg0.text("--->")
#             self.msg1.text("now pick a target point")
#         elif evt.object == self.target:
#             self.targets.append(evt.picked3d)
#             self.source.pickable(True)
#             self.target.pickable(False)
#             self.msg0.text("now pick a source point")
#             self.msg1.text("<---")
#         self.update()
#
#     def on_keypress(self, evt):
#         if evt.keypress == "c":
#             self.sources.clear()
#             self.targets.clear()
#             self.at(0).remove("source_pts")
#             self.at(1).remove("target_pts")
#             self.at(2).remove("warped")
#             self.msg0.text("CLEARED! Pick a point here")
#             self.msg1.text("")
#             self.source.pickable(True)
#             self.target.pickable(False)
#             self.update()
#         elif evt.keypress == "d":
#             n = min(len(self.sources), len(self.targets))
#             self.sources = self.sources[:n-1]
#             self.targets = self.targets[:n-1]
#             self.msg0.text("Last point deleted! Pick a point here")
#             self.msg1.text("")
#             self.source.pickable(True)
#             self.target.pickable(False)
#             self.update()            
################################################################################

settings.default_font = "Calco"

source = Mesh(dataurl+"limb_surface.vtk").color("k5")
source.rotate_y(90).rotate_z(-60).rotate_x(40)

target = Mesh(dataurl+"290.vtk").color("yellow5")
target.rotate_y(-40)

plt = MorphPlotter(source, target, size=(2400, 850), axes=14)
plt.cmap_name = "RdYlBu_r"
plt.show()
plt.close()




"""Interaction mode to fly over a surface.
- Press arrows to move the camera in the plane of the surface.
- "t" and "g" will move the camera up and down along z.
- "x" and "X" will reset the camera to the default position towards +/-x.
- "y" and "Y" will reset the camera to the default position towards +/-y.
- "." and "," will rotate azimuth to the right or left.
- "r" will reset the camera to the default position."""
from vedo import *
from vedo.interactor_modes import FlyOverSurface

settings.enable_default_keyboard_callbacks = False
settings.enable_default_mouse_callbacks = False

surf = ParametricShape("RandomHills").cmap("Spectral")

mode = FlyOverSurface()
txt = Text2D(__doc__, c="k", font="Antares", s=0.8)

plt = Plotter(size=(1200, 600))
plt.user_mode(mode)
plt.show(surf, Axes(surf), txt, elevation=-90, zoom=2, axes=14)
plt.close()



import numpy as np
from vedo import Volume, show

# make up some fake data
X, Y, Z = np.mgrid[:4, :4, :2]
scalar_field = (X - 2) ** 2 + (Y - 2) ** 2 + (Z - 2) ** 2

vol = Volume(scalar_field.astype(int))
spacing = vol.spacing() # get the voxel size
# print("spacing", spacing)
# print('numpy array from Volume:', vol.pointdata)
# print('input_scalars', vol.pointdata['input_scalars'])

# extract a z-slice at index k=1
zslice = vol.zslice(k=1)
zslice.cmap("hot_r").lw(1).alpha(0.9).add_scalarbar3d()
# print("input_scalars", zslice.pointdata["input_scalars"])

# create a set of points at the cell centers and associate the scalar value
# corresponding to the bottom left corner of each voxel
cc = zslice.cell_centers().shift([-spacing[0] / 2, -spacing[1] / 2, 0])
cc.resample_data_from(zslice)

zslice.compute_normals()

zslice2 = zslice.clone()
zslice2.celldata["pixel_value"] = cc.pointdata["input_scalars"]
print(zslice2.celldata["pixel_value"])

lego = vol.legosurface(vmin=0, vmax=10).wireframe()

show(
    [
        (vol, lego, vol.cell_centers()),
        (
            lego,
            cc,
            zslice,
            zslice.labels("id"),
            zslice.labels("cellid"),
        ),
        zslice2,
    ],
    N=3,
    axes=1,
)



"""Probe a Volume with a line and plot the intensity values"""
from vedo import dataurl, Volume, Line, show
from vedo.pyplot import plot

vol = Volume(dataurl + "embryo.slc")
vol.add_scalarbar3d("wild-type mouse embryo", c="k")
vol.scalarbar = vol.scalarbar.clone2d("bottom-right", 0.2)

p1, p2 = (50, 50, 50), (200, 200, 200)
pl = Line(p1, p2, res=100).lw(4)

# Probe the Volume with the line
pl.probe(vol)

# Get the probed values along the line
xvals = pl.points[:, 0]
yvals = pl.pointdata[0]

# Plot the intensity values
fig = plot(
    xvals,
    yvals,
    xtitle=" ",
    ytitle="voxel intensity",
    aspect=16 / 9,
    spline=True,
    lc="r",  # line color
    marker="O",  # marker style
)
fig = fig.shift(0, 25, 0).clone2d()

show(vol, pl, fig, __doc__, axes=14).close()



"""Generate a polygonal Mesh from a contour line"""
from vedo import dataurl, Assembly, Line, show
from vedo.pyplot import histogram

shapes = Assembly(dataurl + "timecourse1d.npy")  # group of lines
shape = shapes[56]  # pick one
cmap = "RdYlBu"

# Generate the Mesh from the line
msh = shape.generate_mesh(invert=True)
msh.smooth()           # make the triangles more uniform
msh.compute_quality()  # add a measure of triangle quality
msh.cmap(cmap)

contour = Line(shape).c("red4").lw(5)
labels = contour.labels("id")

histo = histogram(
    msh.celldata["Quality"],
    xtitle="triangle mesh quality",
    aspect=25/9,
    c=cmap,
).clone2d("bottom-right")

show(contour, labels, msh, histo, __doc__).close()



"""Slice a Volume with an arbitrary plane
hover the plane to get the scalar values"""
from vedo import dataurl, precision, Sphere, Volume, Plotter

def func(evt):
    if not evt.object:
        return
    pid = evt.object.closest_point(evt.picked3d, return_point_id=True)
    txt = f"Probing:\n{precision(evt.object.picked3d, 3)}\nvalue = {arr[pid]}"

    pts = evt.object.points
    sph = Sphere(pts[pid]).c('orange7').pickable(False)
    fp = sph.flagpole(txt, s=7, offset=(-150,15), font=2).follow_camera()
    # remove old and add the two new objects
    plt.remove('Sphere', 'FlagPole').add(sph, fp).render()

vol = Volume(dataurl+'embryo.slc')
vol.cmap('white').alpha([0,0,0.8]).pickable(False)

vslice = vol.slice_plane(vol.center(), [0,1,1])
vslice.cmap('Purples_r').add_scalarbar('Slice', c='w')
arr = vslice.pointdata[0] # retrieve vertex array data
print("slice shape :", vslice.metadata["shape"])
print("slice bounds:", vslice.metadata["original_bounds"])

plt = Plotter(axes=9, bg='k', bg2='bb')
plt.add_callback('as my mouse moves please call', func) # be kind to vedo ;)
plt.show(vol, vslice, __doc__)
plt.close()




from vedo import *

one = Mesh(dataurl+"bunny.obj", c="green")
two = Mesh(dataurl+"bunny.obj", c="red").shift(0.3,0,0)

one.add_shadow("z", -0.1)
two.add_shadow("z", -0.1)

show(one, two, axes=1).close()


"""Threshold a TetMesh with a scalar array"""
from vedo import *

tetm = TetMesh(dataurl + "limb.vtu")

# Threshold the tetrahedral mesh for values in the range:
tetm.threshold(above=0.9, below=1)

tetm.celldata.select("chem_0").cmap("Accent")
tetm.add_scalarbar3d("chem_0  expression levels", c="k", italic=1)

# Make a 2D clone of the 3D scalarbar and place it to the right:
tetm.scalarbar = tetm.scalarbar.clone2d("center-right", size=0.2)

show(tetm, __doc__, axes=1).close()



from vedo import VedoLogo, settings

settings.use_parallel_projection = 1
settings.screenshot_transparent_background = 0

vl = VedoLogo(frame=False, c='k')
vl.show(size=(340*3,115*3), zoom="tight")
# vl.screenshot("logo_vedo_simple.png")



from vedo import settings, show
from vedo.pyplot import pie_chart

settings.default_font = "Komika"

title     = "A pie chart plot"
fractions = [0.1, 0.2, 0.3, 0.1, 0.3]
colors    = [ 1,   2,   3,   4, 'white']
labels    = ["stuff_1 ", "stuff_2 ", "comp^A ", "comp^B ", ""]

pc = pie_chart(fractions, c=colors, labels=labels, title=title)
pc2d = pc.clone2d("top-left", size=0.975, ontop=False)

show(pc2d).close()



import numpy as np
import vedo

N = np.arange(24).reshape([2, 3, 4])
cubes = []
texts = []
positions = []
xs, ys, zs = N.shape
for x in range(xs):
    for y in range(ys):
        for z in range(zs):
            pos = (x, y, z)
            val = N[x, y, z]
            cubes.append(vedo.Cube(pos=pos, side=0.6, alpha=0.1))
            positions.append(pos)

pts = vedo.Points(positions)
labs= pts.labels2d(font='Quikhand', scale=2, justify="top-center", c="red4")
vedo.show(cubes, labs, axes=4).close()

################################################################### (BUG)
texts = []
xs, ys, zs = [2, 1, 2]
for x in range(xs):
    for y in range(ys):
        for z in range(zs):
            pos = (x, y, z)
            txt = vedo.Text3D(f"{pos}", pos, s=0.05, justify='centered', c='r5')
            txt.rotate_x(0.00001)
            txt.shift(0.00001, 0.00001, 0.00001) # same as rotate_x
            texts.append(txt.follow_camera())

vedo.show(texts, axes=1)



from vedo import *

def callb(evt):
    msh = evt.object
    if not msh:
        return
    pt = evt.picked3d
    idcell = msh.closest_point(pt, return_cell_id=True)
    # msh.cellcolors[idcell] = [255,0,0,255] # red, opaque
    cols = msh.cellcolors.copy()
    cols[idcell] = [0,255,0,255] # green, opaque
    msh.cellcolors = cols
    plt.render()
    
m = Mesh(dataurl + "290.vtk")
m.decimate().smooth().compute_normals()
m.compute_quality().cmap("Blues", on="cells")

print(m.cellcolors)

plt = Plotter()
plt.add_callback("mouse click", callb)
plt.show(m, m.labels("cellid"))
plt.close()




"""
Perform other simple mathematical operation between 3D Volumes.
Possible operations are: +, -, /, 1/x, sin, cos, exp, log, abs, **2, sqrt,
  min, max, atan, atan2, median, mag, dot, gradient, divergence, laplacian.
Alphas defines the opacity transfer function in the scalar range.
"""
from vedo import *
printc(__doc__)

plt = Plotter(N=4)

v0 = Volume(dataurl+'embryo.slc').cmap(0).add_scalarbar3d()
v0.scalarbar = v0.scalarbar.clone2d("center-right", 0.1) # substitute to a 2d scalarbar
plt.at(0).show("original", v0)

v1 = v0.clone().operation("gradient").operation("mag").add_scalarbar3d()
v1.scalarbar = v1.scalarbar.clone2d("center-right", 0.1)
plt.at(1).show("gradient", v1)

v2 = v0.clone().operation("median").cmap(4).add_scalarbar3d()
v2.scalarbar = v2.scalarbar.clone2d("center-right", 0.1)
plt.at(2).show("median", v2)

v3 = v0.clone().operation("dot", v0).cmap(7).add_scalarbar3d()
v3.scalarbar = v3.scalarbar.clone2d("center-right", 0.1)
plt.at(3).show("dot(v0,v0)", v3, zoom=1.3)

plt.interactive().close()


####################################################################################
#Start with creating a masked Volume then compute its gradient and probe 2 points
msh = Ellipsoid()

vol = msh.signed_distance(dims=[20, 20, 20])
vol.threshold(above=0.0, replace=0.0)  # replacing all values outside to 0
vol.cmap("blue").alpha([0.9, 0.0]).alpha_unit(0.1).add_scalarbar3d()
vol.scalarbar = vol.scalarbar.clone2d("center-right", 0.15)

vgrad = vol.operation("gradient")
printc(vgrad.pointdata, c='g')

grd = vgrad.pointdata['ImageScalarsGradient']
pts = vol.points  # coords as numpy array
arrs = Arrows(pts, pts + grd*0.1, c="jet")

pts_probes = [[0.2,0.5,0.5], [0.2,0.3,0.4]]
vpts_probes = Points(pts_probes).probe(vgrad)
vects = vpts_probes.pointdata['ImageScalarsGradient']

arrs_pts_probe = Arrows(pts_probes, pts_probes + vects, c='black')

plt = Plotter(axes=1, N=2)
plt.at(0).show("A masked Volume", vol)
plt.at(1).show("..compute its gradient and probe 2 points", arrs, arrs_pts_probe)
plt.interactive().close()



"""Build a TetMesh (tetrahedral mesh)
by manually defining vertices and cells"""
from vedo import *

points = [ 
    (0, 0, 0), # first tet
    (1, 0, 0),
    (1, 1, 0),
    (0, 1, 2),
    (3, 3, 3), # second tet
    (4, 3, 3),
    (4, 4, 3),
    (3, 4, 4),
    (2, 5, 3), # third tet
    (3, 5, 3),
    (3, 6, 3),
    (2, 5, 5),
]

# Cells are defined by a list of 4 vertex indices
# note that "cells" and "tetrahedrons" are the same thing
tets = [[0,1,2,3], [4,5,6,7], [8,9,10,11]]

# Define a scalar value for each cell we have created
values = np.array([10.0, 20.0, 30.0])

# Create the TeTMesh object and assign any number of data arrays to it
tm = TetMesh([points, tets])
tm.celldata["myscalar1"] = values
tm.celldata["myscalar2"] = -values / 10
tm.pointdata["myvector"] = np.random.rand(tm.npoints)
# ...

print(tm)

tm.celldata.select("myscalar2").cmap('jet').add_scalarbar()
# tm.color('green') # or set a single color

# Create labels for the vertices
labels = tm.labels2d('id', scale=2)

show(tm, labels, __doc__, axes=1).close()



"""Add an observer to specific objects in a scene"""
from vedo import *

# -----------------------
def func(obj, name=None):
    printc("Plotter callback", c="m")

# -----------------------
def ftxt(obj, ename):
    printc("Text2D callback", obj.__class__.__name__, ename, c="y")
    obj.color(np.random.rand() * 10)

# -----------------------
def fmsh(obj, ename):
    printc("Mesh callback", obj.__class__.__name__, ename, c="b")
    msh.color(np.random.rand() * 10)


msh = Mesh(dataurl + "spider.ply")
cid2 = msh.add_observer("pick", fmsh)

txt = Text2D("CLICK ME", pos="bottom-center", s=3, bg="yellow5").pickable()
cid1 = txt.add_observer("pick", ftxt)

plt = Plotter()
# plt.add_observer("mouse click", func)  ### SAME AS:
# plt.add_callback("mouse click", func, enable_picking=False)
plt.show(txt, msh, __doc__).close()


"""Explore RGB and HSV color spaces"""
from vedo import *
from vedo.colors import rgb2hsv, hsv2rgb, rgb2hex

def update_txt(rgb, hsv):
    RGB = np.round(np.array(rgb)*255).astype(int)
    HEX = rgb2hex(rgb)
    name = get_color_name(rgb)
    tx1.text(f"RGB: {precision(rgb, 3)}\n     {RGB}\nHEX: {HEX}")
    tx2.text(f"HSV: {precision(hsv, 3)}\n ~ {name}")
    box.color(rgb)

def func_rgb(w, e):
    rgb = slr.value, slg.value, slb.value
    hsv = rgb2hsv(rgb)
    slh.value, sls.value, slv.value = hsv
    update_txt(rgb, hsv)

def func_hsv(w, e):
    hsv = slh.value, sls.value, slv.value
    rgb = hsv2rgb(hsv)
    slr.value, slg.value, slb.value = rgb
    update_txt(rgb, hsv)

box = Cube().linewidth(2).color([0.5, 0.5, 0.5]).lighting("off")
tx1 = Text2D(font="Calco", s=1.4, pos="top-left",  bg="k5").text(__doc__)
tx2 = Text2D(font="Calco", s=1.4, pos="top-right", bg="k5")

plt = Plotter()

slr = plt.add_slider(func_rgb, 0, 1, value=0.5, show_value=False, c="r3", pos=((0.05,0.18),(0.40,0.18)))
slg = plt.add_slider(func_rgb, 0, 1, value=0.5, show_value=False, c="g3", pos=((0.05,0.12),(0.40,0.12)))
slb = plt.add_slider(func_rgb, 0, 1, value=0.5, show_value=False, c="b3", pos=((0.05,0.06),(0.40,0.06)), title="RGB")

slh = plt.add_slider(func_hsv, 0, 1, value=0.5, show_value=False, c="k1", pos=((0.60,0.18),(0.95,0.18)))
sls = plt.add_slider(func_hsv, 0, 1, value=0.0, show_value=False, c="k1", pos=((0.60,0.12),(0.95,0.12)))
slv = plt.add_slider(func_hsv, 0, 1, value=0.5, show_value=False, c="k1", pos=((0.60,0.06),(0.95,0.06)), title="HSV")

plt.show(box, tx1, tx2, viewup="z").close()




from vedo import Volume, Text3D, show, dataurl, settings

settings.use_parallel_projection = True

vol = Volume(dataurl+"embryo.slc")
vol.mode(0).c('b9').alpha_unit(1)

t = Text3D("Sharpe\n~~~Lab", s=40, font="Spears", vspacing=1.4)
t.c('k9').rotate_x(90).pos(200,150,70)

cam = dict(
    position=(227.421, -911.244, 192.438),
    focal_point=(217.166, 126.841, 116.242),
    viewup=(0, 0, 1),
    parallel_scale=110,
    clipping_range=(754.990, 1403.38),
)

plt = show(vol, t, size=(700,400), camera=cam, bg='bb')
settings.screenshot_transparent_background = 1
plt.screenshot("logo.png")




"""Create an animated logo"""
from vedo import *
from vedo.pyplot import histogram

settings.use_parallel_projection = True

exa = Polygon().scale(4.1).pos(5.25, 4.8, 0).off()
his = histogram([-1, 1], [-1, 1], mode='hex').unpack()

exah, cmh = [], []
for h in his:
    cm = h.center_of_mass()
    if exa.is_inside(cm):
        h.c('green').shrink(0.9).add_shadow(plane='z', point=-.4)
        exah.append(h)
        cmh.append(cm)

v1 = vector(9.4, 5.2, 0)
v2 = vector(9.4, 2.7, 0)
t1 = Text3D("EMBL",  v1, c="k", font="VTK", s=1.5, depth=0)
t2 = Text3D("European Molecular\nBiology Laboratory", v2, font="VTK", vspacing=1.75, c="dg", s=0.6)

plt = show(exa, exah, t1, t2, axes=0, interactive=0, elevation=-50)
for ti in reversed(range(100)):
    t = ti / 100.
    for j, h in enumerate(exah):
        cx, cy, _ = cmh[j] - [4,5,0]
        x = t*-4+(1-t)*6
        g = exp(-(cx-x)**2/.5)*2
        h.z(g)
        t1.pos([sin(t)*-10, 0, -0.41] + v1).alpha((1-t)**2)
        t2.pos([sin(t)*-15, 0, -0.41] + v2).alpha((1-t)**4)
        exah[13].c('red')
    plt.show(exa, exah, t1, t2, resetcam=0, elevation=t, azimuth=-0.02)

plt.interactive()



"""Setting illumination properties:
ambient, diffuse, specular power and color."""
from vedo import Plotter, Mesh, dataurl

ambient = 0.1
diffuse = 0
specular = 0
specular_power = 20
specular_color = "white"

apple = Mesh(dataurl + "apple.ply")
apple.flat().c("gold")

plt = Plotter(axes=1, bg='black', bg2='white')

for i in range(8):
    x = (i % 4) * 2.2
    y = int(i < 4) * 3
    apple_copy = apple.clone().pos(x, y)

    # modify the default with specific values
    apple_copy.lighting(
        "default", ambient, diffuse, 
        specular, specular_power, specular_color
    )
    plt += apple_copy

    ambient += 0.125
    diffuse += 0.125
    specular+= 0.125

plt += __doc__
plt.show().close()



"""Interpolate a 2D surface through a set of points"""
import numpy as np
from scipy.spatial import distance_matrix
from vedo import Grid, Points, Lines, show
np.random.seed(1)

def harmonic_shepard(pts, vals, radius):
    dists = distance_matrix(pts, pts) + radius
    rdists = 1.0 / dists
    sum_vals = np.sum(rdists * vals, axis=1)
    return sum_vals / np.sum(rdists, axis=1)

# Create a grid of points
surf = Grid(res=[25,25])

# Pick n random points on the surface
ids = np.random.randint(0, surf.npoints, 10)
pts = surf.coordinates[ids]

# Create a set of random scalars
scals1 = np.random.randn(10) * 0.1 

ptss1 = pts.copy()
ptss1[:,2] = scals1   # assign scalars as z-coords
pts1 = Points(ptss1).color("red5").point_size(15)

# Compute an interpolated (smoothed) set of scalars
scals2 = harmonic_shepard(pts, scals1, radius=0.1)
ptss2 = pts.copy()
ptss2[:,2] = scals2
pts2 = Points(ptss2).color("purple5").point_size(15)

# Warp the surface to match the interpolated points
ptsource, pttarget = [], []
for pt in pts2.coordinates:
    pt_surf = surf.closest_point(pt)
    ptsource.append(pt_surf)
    pttarget.append(pt)
warped = surf.warp(ptsource, pttarget, mode='2d')
warped.color("b4").lc('lightblue').wireframe(False).lw(1)

lines = Lines(pts1, pts2, lw=2)

show(pts1, pts2, lines, warped, __doc__, axes=1, viewup="z").close()




"""Customizing a legend box"""
from vedo import *

s = Sphere()
c = Cube().x(2)
e = Ellipsoid().x(4)

h = Hyperboloid().x(6)
h.legend('The description for\nthis one is quite long')

lbox = LegendBox([s,c,e,h], width=0.3, height=0.4, markers='s')
lbox.font("Kanopus")

show(s, c, e, h, lbox, __doc__,
     axes=1, bg='lightyellow', bg2='white', size=(1200,800), viewup='z'
).close()




import sys
from PyQt5 import Qt
from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from vedo import Plotter, Cone, printc, settings


class MainWindow(Qt.QMainWindow):
 
    def __init__(self, parent=None):

        Qt.QMainWindow.__init__(self, parent)
        self.frame = Qt.QFrame()
        self.layout = Qt.QVBoxLayout()
        self.vtkWidget = QVTKRenderWindowInteractor(self.frame)

        # Create renderer and add the vedo objects and callbacks
        self.plt = Plotter(qt_widget=self.vtkWidget)
        self.plt += Cone().rotate_x(20)

        self.button = self.plt.add_button(
            self.buttonfunc,
            pos=(0.7, 0.05),  # x,y fraction from bottom left corner
            states=["click to green"],  # text for each state
            c=["w"],  # font color for each state
            bc=["dg"],  # background color for each state
            font="courier",  # font type
            size=25,  # font size
            bold=True,  # bold font
            italic=False,  # non-italic font style
        )

        self.plt.show()  # <--- show the vedo rendering

        # Set-up the rest of the Qt window
        button = Qt.QPushButton("My Button makes the cone red")
        button.setToolTip("This is an example button")
        button.clicked.connect(self.onClick)
        self.layout.addWidget(self.vtkWidget)
        self.layout.addWidget(button)
        self.frame.setLayout(self.layout)
        self.setCentralWidget(self.frame)
        self.show()  # <--- show the Qt Window

    def buttonfunc(self, obj, ename):
        print("btn is clicked...")
        self.plt.objects[0].color("green5").rotate_z(40)
        
    @Qt.pyqtSlot()
    def onClick(self):
        printc("..calling onClick")
        self.plt.objects[0].color("red5").rotate_z(40)
        self.plt.render()


if __name__ == "__main__":
    if settings.dry_run_mode:
        sys.exit()
    app = Qt.QApplication(sys.argv)
    window = MainWindow()
    app.exec_()


from vedo import Latex, show
from vedo.pyplot import histogram
import numpy as np

N = 2000
x = np.random.randn(N) * 1.0
y = np.random.randn(N) * 1.5

# hexagonal binned histogram:
histo = histogram(x, y, bins=10, mode="hexbin", fill=True, cmap="terrain")
# add a formula:
f  = r"f(x, y)=A \exp \left(-\left(\frac{\left(x-x_{o}\right)^{2}}"
f += r"{2 \sigma_{x}^{2}}+\frac{\left(y-y_{o}\right)^{2}}"
f += r"{2 \sigma_{y}^{2}}\right)\right)"

formula = Latex(f, c="k", s=1.5)
formula.rotate_x(90).rotate_z(90).pos([-4, -5, 2])

show(histo, formula, axes=1, viewup="z").close()



"""Cut a TetMesh with an arbitrary polygonal Mesh"""
from vedo import *

sphere = Sphere(r=500).x(400)
sphere.color('green5', 0.2).wireframe()

tmesh = TetMesh(dataurl + 'limb.vtu')
print(tmesh)

ugrid = tmesh.cut_with_mesh(sphere, invert=True).cmap("Reds_r")
print(ugrid)

# We may cast the output to a new TetMesh:
# tmesh_cut = TetMesh(ugrid)
# print(tmesh_cut)

show(ugrid, sphere, __doc__, axes=1).close()



# 2D Fast Fourier Transform of a image
from vedo import Image, show

url = 'https://vedo.embl.es/examples/data/images/dog.jpg'

img = Image(url).resize([200,None])  # resize so that x has 200 pixels, but keep y aspect-ratio
img_fft = img.fft(logscale=12)
img_fft = img_fft.tomesh().cmap('Set1',"RGBA").add_scalarbar("12 log(fft)")  # optional step

show([
      [img,    f"Original image\n{url[-40:]}"],
      [img_fft, "2D Fast Fourier Transform"],
      [img.fft(mode='complex').rfft(), "Reversed FFT"],
     ],
     N=3, bg='gray7', axes=1,
).close()





"""Start typing a color name for the mesh.
E.g.: pink4
(Press 'Esc' to exit)"""
from vedo import settings, dataurl, get_color_name
from vedo import Plotter, Mesh, Text2D


def kfunc(evt):
    key = evt.keypress.lower()
    field_txt = field.text().strip() # strip leading/trailing spaces

    if key == "backspace" and field_txt:
        key = ""
        field_txt = field_txt[:-1]
    elif key == "escape":
        plt.close()
        return
    elif len(key) > 1:
        return

    color_name = field_txt + key
    field.text(f"{color_name:^12}").frame(color_name, lw=8)
    mesh.color(color_name)
    msg.text(get_color_name(color_name))
    plt.render()


settings["enable_default_keyboard_callbacks"] = False

mesh = Mesh(dataurl+"magnolia.vtk").color("black").flat()

field = Text2D("black", pos="bottom-center",s=3, font="Meson", bg="k2", c="w", alpha=1)
msg = Text2D(pos="top-right", s=2, font="Quikhand", c="k1", bg="k7", alpha=1)

plt = Plotter()
plt.add_callback("key press", kfunc)
plt.show(mesh, field, msg, __doc__).close()



import numpy as np
import napari
import vedo

print("\nSEE ALSO https://github.com/jo-mueller/napari-vedo-bridge")

# Load the surface, triangulate just in case, and compute vertex normals
surf = vedo.Mesh(vedo.dataurl+"beethoven.ply").triangulate().compute_normals()
surf.rotate_x(180).rotate_y(60)

vertices = surf.vertices
faces    = np.array(surf.cells)
normals  = surf.vertex_normals
# generate vertex values by projecting normals on a "lighting vector"
values   = np.dot(normals, [-1, 1, 1])

# create an empty viewer
viewer = napari.Viewer()

# add the surface
viewer.add_surface((vertices, faces, values), opacity=0.8)
viewer.add_points(vertices, size=0.05, face_color='pink')

# turn on 3D rendering
viewer.dims.ndisplay = 3
napari.run()


"""Create a button using an Image icon to show its state"""
from vedo import *

def button_func(widget, evtname):
    print("button_func called")
    cone.color(button.state)
    if button.state:
        plt.background("black")
    else:
        plt.background("white")

def on_mouse_click(event):
    if event.object:
        print("on_mouse_click", event)
        cone.color(button.state)

# Create a cone
cone = Cone().color(0)

# Create a plotter
plt = Plotter(bg='w', axes=1)
plt.add_callback('mouse click', on_mouse_click)

plt.add(cone, __doc__)

# Create a button widget
img0 = Image(dataurl+"images/play-button.png")
img1 = Image(dataurl+"images/power-on.png")

button = ButtonWidget(
    button_func, 
    # states=["State 0", "State 1"], 
    states=[img0, img1],
    c=["red4", "blue4"],
    bc=("k9", "k5"),
    size=100,
    plotter=plt,
)
button.pos([0,0]).enable()

plt.show(elevation=-40)



"""Create three checkbox buttons to toggle objects on/off."""
from vedo import Mesh, Plotter, dataurl

s1 = Mesh(dataurl+"bunny.obj").normalize().x(0).color("p5")
s2 = Mesh(dataurl+"teapot.vtk").normalize().x(3).rotate_x(-90).color("y5")
s3 = Mesh(dataurl+"mug.ply").normalize().x(6).color("r5")


def func1(b, evt):
    s1.toggle()  # toggle visibility
    b.switch()

def func2(b, evt):
    s2.toggle()
    b.switch()

def func3(b, evt):
    s3.toggle()
    b.switch()

def func4(_, evt):
    [s.toggle() for s in (s1,s2,s3)]
    [b.switch() for b in plt.buttons]

plt = Plotter(axes=1, size=(1000,500))
plt.add_hint(s1, "A Bunny",  size=42) # shows a label when hovering on the object
plt.add_hint(s2, "A Teapot", size=42)
plt.add_hint(s3, "A Mug",    size=42)

plt.add_button(func1, pos=(0.4,0.15), size=42, states=["", ""], bc=["p5", "k7"])
plt.add_button(func2, pos=(0.5,0.15), size=42, states=["", ""], bc=["y5", "k7"])
plt.add_button(func3, pos=(0.6,0.15), size=42, states=["", ""], bc=["r5", "k7"])
plt.add_button(func4, pos=(0.9,0.15), size=42, states=["flip","flip"], bc=["k4","k5"],
               c=["k5","k4"], font="Cartoons123")

plt.show(s1, s2, s3, __doc__, zoom=1.8).close()



"""Load a tetrahedral mesh and show it in different styles."""
from vedo import TetMesh, show, dataurl

# Load a tetrahedral mesh from file
tetm = TetMesh(dataurl + 'limb.vtu')
tetm.compute_cell_size()
print(tetm)

# Assign a color to each tetrahedron based on the value of "chem_0"
tetm.celldata.select('chem_0').cmap("Blues_r").add_scalarbar()

# Make a copy of tetm and shrink the tets
msh = tetm.clone().shrink(0.5).tomesh().add_scalarbar()

# Show the two meshes side by side with comments
show([(tetm, __doc__), (msh, "..shrunk tetrahedra")], N=2).close()




"""Andrews curves for the Iris dataset."""
import numpy as np
from sklearn import datasets
from vedo import *
from vedo.pyplot import Figure, plot

iris = datasets.load_iris()  # loading iris data set

AC = andrews_curves(iris.data)
theta = np.linspace(-np.pi, np.pi, 100)

settings.remember_last_figure_format = True
fig = Figure(
    xlim=(-np.pi, np.pi),
    ylim=(-2, 16),
    padding=0,
    axes=dict(htitle="", axes_linewidth=2, xyframe_line=0),
)

setosas = []
for r in AC[:20]:  # setosa
    p = pol2cart(r, theta).T
    fig += plot(theta, r, c="red5")
    setosas.append(Line(p))
setosas = merge(setosas).lw(3).c("red5")

versicolors = []
for r in AC[50:70]:  # versicolor
    p = pol2cart(r, theta).T
    fig += plot(theta, r, c="blue5")
    versicolors.append(Line(p))
versicolors = merge(versicolors).lw(3).c("blue5")

virginicas = []
for r in AC[100:120]:  # virginica
    p = pol2cart(r, theta).T
    fig += plot(theta, r, c="green5")
    virginicas.append(Line(p))
virginicas = merge(virginicas).lw(3).c("green5")

fig = fig.clone2d(size=0.75)
show(setosas, versicolors, virginicas, fig, __doc__, axes=12)



"""A RectilinearGrid is a dataset where edges are parallel to the coordinate axes.
It can be thought of as a tessellation of a box in 3D space, similar to a Volume
except that the cells are not necessarily cubes, but they can have different
lengths along each axis."""

from vedo import *

xcoords = 7 + np.sqrt(np.arange(0,900,25))
ycoords = np.arange(0, 20)
zcoords = np.arange(0, 20)

rgrid = RectilinearGrid([xcoords, ycoords, zcoords])

print(rgrid)
# print(rgrid.x_coordinates().shape)
# print(rgrid.has_blank_points())
# print(rgrid.compute_structured_coords([20,10,11]))

msh = rgrid.tomesh().lw(1)
show(msh, __doc__, axes=1, viewup="z")



"""Create a scatter plot to overlay
three different distributions"""
from vedo import *
from numpy.random import randn

### first cloud in blue, place it at z=0:
x = randn(2000) * 3
y = randn(2000) * 2
xy = np.c_[x, y]
pts1 = Points(xy).z(0.0).color("blue5",0.5)
bra1 = Brace([-7, -8], [7, -8], comment="whole population", s=0.4).c("blue5")

### second cloud in red
x = randn(1200) + 4
y = randn(1200) + 2
xy = np.c_[x, y]
pts2 = Points(xy).z(0.1).color("red",0.5)
bra2 = Brace(
    [8, 2, 0.3],
    [6, 5, 0.3],
    comment="red zone",
    angle=180,
    justify="bottom-center",
    c="red5",
)

### third cloud with a black marker
x = randn(20) + 4
y = randn(20) - 4
mark = Marker("*", s=0.25)
pts3 = Glyph(xy, mark).z(0.2).color("red5",0.5)
bra3 = Brace([8, -6], [8, -2], comment="my stars").z(0.3)

# some text message
msg = Text3D("preliminary\nresults!", font="Quikhand", s=1.5).c("black")
msg.rotate_z(20).pos(-10, 3, 0.2)

show(
    pts1, pts2, pts3, msg, bra1, bra2, bra3, __doc__,
    axes=1, zoom=1.2, mode="image",
).close()



"""Stream tubes originating from a probing grid of points.
Data is from CFD analysis of airflow in an office with
ventilation and a burning cigarette."""
from vedo import *
from off_furniture import furniture

fpath = download(dataurl + 'office.binary.vtk')
sgrid = loadStructuredGrid(fpath)
ugrid = UnstructuredGrid(sgrid) # convert to unstructured grid which vedo supports

# Create a grid of points and use it as integration seeds
seeds = Grid(res=[2,3], c="white").rotate_y(90).pos(2,2,1)

streamlines = ugrid.compute_streamlines(seeds, initial_step_size=0.01, max_propagation=15)
streamlines.cmap("Reds").add_scalarbar3d(c='white')
streamlines.scalarbar = streamlines.scalarbar.clone2d("center-right", size=0.15)
print(streamlines)

show(streamlines, seeds, furniture(), __doc__, axes=1, bg='bb').close()



"""Isosurface extraction from a volume dataset
with discrete values (labels)."""
from vedo import *

settings.default_font = "Antares"

lut = build_lut(
    [
        [0, "lightyellow"],
        [1, "red8"],
        [2, "blue"],
        [3, "yellow"],
        [4, "orange"],
        [5, "cyan"],
        [6, "magenta"],
        [7, "white"],
        [8, "pink"],
        [9, "brown"],
        [10, "lightblue"],
        [11, "lightgreen"],
        [12, "red4"],
    ],
    interpolate=False,
)

# This dataset is a 3D volume of 64x64x64 voxels containing 12 "blobs"
blobs = Volume(dataurl + "blobs.vti")
box = blobs.box()

isovalues = list(range(1, 13))
iso_discrete = blobs.isosurface_discrete(
    isovalues,
    background_label=0,
    internal_boundaries=True,
    nsmooth=10,
)
iso_discrete.cmap(lut)

separate_blobs = []
txt_vols = []
for i in isovalues:
    b = iso_discrete.clone().threshold(0, i - 0.5, i + 0.5, on="cells")
    b.cap().clean().compute_normals()
    b.color(i).alpha(0.1).wireframe().lighting("off")
    v = b.volume() / 1e3
    cm = b.center_of_mass()
    txt = Text3D(f"blob {i}\nvol={v:.2f}", c=i, justify="center", pos=cm)
    txt_vols.append(txt)
    separate_blobs.append(b)

show([[iso_discrete, box, __doc__], [separate_blobs, txt_vols, box]], N=2, axes=1)




from vedo import Volume, dataurl
from vedo.applications import RayCastPlotter

# Load Volume data
embryo = Volume(dataurl + "embryo.slc")
embryo.mode(1).cmap("jet")  # change visual properties

# Create a Plotter instance and show
plt = RayCastPlotter(embryo, bg='black', bg2='blackboard', axes=7)
plt.show(viewup="z")
plt.close()



"""Convert a vedo mesh to a madcad mesh and vice versa"""
# See https://pymadcad.readthedocs.io/en/latest/index.html
import vedo
import madcad

mesh = vedo.Mesh(vedo.dataurl+"bunny.obj")
mesh.compute_normals()

############################################################
madcad_mesh = vedo.utils.vedo2madcad(mesh)

madcad.thicken(madcad_mesh, thickness=0.1)
if vedo.settings.dry_run_mode == 0:
    madcad.show([madcad_mesh])


#############################################################
vedo_mesh = vedo.utils.madcad2vedo(madcad_mesh)

verts = vedo_mesh.vertices
norms = vedo_mesh.pointdata["Normals"]
arrs = vedo.Arrows(verts, verts + 0.005 * norms)
vedo.show(mesh, arrs, __doc__, axes=1).close()



"""Include background images in the rendering scene
(generated by matplotlib)"""
import matplotlib.pyplot as plt
from vedo import *

tmsh = TetMesh(dataurl + "limb.vtu")
msh = tmsh.tomesh().shrink(0.8)

# Create a histogram with matplotlib
fig = plt.figure()
plt.hist(msh.celldata["chem_0"], log=True)
plt.title(r"$\mathrm{Matplotlib\ Histogram\ of\ log(chem_0)}$")

# pic1 = Image(fig).clone2d("top-right", 0.5).alpha(0.8)
pic2 = Image(dataurl + "images/embryo.jpg").clone2d("bottom-right")

show(msh, fig, pic2, __doc__, bg="lightgrey", zoom=1.2, axes=1)



## This contains the script snippets that come with the documetation for testing
import numpy as np
from vedo import *
from vedo.pyplot import plot
import vedo

doshow = 1

##################################################################### addons.py
box = Box(pos=(1,2,3), length=8, width=9, height=7).alpha(0.1)
axs = Axes(box, c='k')  # returns Assembly object
for a in axs.unpack():
    print(a.name)
if doshow:
    show(box, axs).close()

######################################################
print("Test 1")
b = Box(pos=(0, 0, 0), length=80, width=90, height=70).alpha(0.1)
if doshow:
    show(
        b,
        axes={
            "xtitle": "Some long variable [a.u.]",
            "number_of_divisions": 4,
            # ...
        },
    ).close()

##################################################################### base.py
print("Test 2")
c1 = Cube()
c2 = c1.clone().c('violet').alpha(0.5) # copy of c1
v = vector(0.2,1,0)
p = vector(1,0,0)  # axis passes through this point
c2.rotate(90, axis=v, point=p)
l = Line(-v+p, v+p).lw(3).c('red')
if doshow:
    show(c1, l, c2, axes=1).close()


######################################################
print("Test 3")
objs = []
for i in range(-5, 5):
    p = [i/3, i/2, i]
    v = vector(i/10, i/20, 1)
    c = Circle(r=i/5+1.2).pos(p).lw(3)
    objs += [c, Arrow(p,p+v)]
if doshow:
    show(objs, axes=1).close()


######################################################
print("Test 4")
c1 = Cube()
c2 = c1.clone().c('violet').alpha(0.5) # copy of c1
v = vector(0.2,1,0)
p = vector(1,0,0)  # axis passes through this point
c2.rotate(90, axis=v, point=p)
# get the inverse of the current transformation
T = c2.transform.compute_inverse()
c2.apply_transform(T)  # put back c2 in place
l = Line(p-v, p+v).lw(3).c('red')
if doshow:
    show(c1.wireframe().lw(3), l, c2, axes=1).close()



######################################################
print("Test 5")
tetmesh = TetMesh(dataurl+'limb_ugrid.vtk')
tetmesh.color('rainbow')
cu = Cube(side=500).x(500) # any Mesh works
tetmesh.cut_with_box(cu)
if doshow:
    show(axes=1).close()

##################################################################### mesh.py
print("Test 6")
s = Sphere().crop(right=0.3, left=0.1)
if doshow:
    show(s).close()

######################################################
print("Test 7")
c1 = Cylinder(pos=(0,0,0), r=2, height=3, axis=(1,.0,0), alpha=.1).triangulate()
c2 = Cylinder(pos=(0,0,2), r=1, height=2, axis=(0,.3,1), alpha=.1).triangulate()
intersect = c1.intersect_with(c2).join(reset=True)
spline = Spline(intersect).c('blue').lw(5)
if doshow:
    show(c1, c2, spline, intersect.labels('id'), axes=1).close()


######################################################
print("Test 8")
grid = Grid()#.triangulate()
circle = Circle(r=0.3, res=24).pos(0.11,0.12)
line = Line(circle, closed=True, lw=4, c='r4')
# grid.imprint(line)
if doshow:
    show(grid, line, axes=1).close()


##################################################################### Image.py
print("Test 9")
if doshow:
    pic = Image(dataurl+'dog.jpg').pad()
    pic.append([pic,pic,pic], axis='y')
    pic.append([pic,pic,pic,pic], axis='x')
    pic.show(axes=1).close()


######################################################
print("Test 10")
if doshow:
    p = vedo.Image(vedo.dataurl+'images/dog.jpg').bw()
    pe = p.clone().enhance()
    show(p, pe, N=2, mode='image', zoom='tight').close()



######################################################
print("Test 11")
if doshow:
    pic1 = Image("https://aws.glamour.es/prod/designs/v1/assets/620x459/547577.jpg")
    pic2 = pic1.clone().invert()
    pic3 = pic1.clone().binarize()
    show(pic1, pic2, pic3, N=3, bg="blue9").close()



######################################################
print("Test 12")
if doshow:
    pic = vedo.Image(vedo.dataurl+"images/dog.jpg")
    pic.add_rectangle([100,300], [100,200], c='green4', alpha=0.7)
    pic.add_line([100,100],[400,500], lw=2, alpha=1)
    pic.add_triangle([250,300], [100,300], [200,400])
    show(pic, axes=1).close()


##################################################################### plotter.py
print("Test 13")
cone = Cone()
if doshow:
    cone.show(axes=1).fly_to([1,0,0])
    cone.show().close()

######################################################
print("Test 14")
settings.use_parallel_projection = True # or else it doesnt make sense!
cube = Cube().alpha(0.2)
plt = Plotter(size=(900,600), axes=dict(xtitle='x (um)'))
if doshow:
    plt.add_scale_indicator(units='um', c='blue4')
    plt.show(cube, "Scale indicator with units").close()
settings.use_parallel_projection = False

######################################################
print("Test 15")
def func(evt): # called every time the mouse moves
    # evt is a dotted dictionary
    if not evt.actor:
        return  # no hit, return
    print("point coords =", evt.picked3d)

elli = Ellipsoid()
plt = Plotter(axes=1)
plt.add_callback('mouse hovering', func)
if doshow:
    plt.show(elli).close()


##################################################################### pointcloud.py
# print("Test 16")
# s = Ellipsoid().rotate_y(30)
# #Camera options: pos, focal_point, viewup, distance,
# # clippingRange, parallelScale, thickness, viewAngle
# camopts = dict(pos=(0,0,25), focal_point=(0,0,0))
# if doshow:
#     show(s, camera=camopts, offscreen=True).close()
#     m = s.visible_points()
#     #print('visible pts:', m.points()) # numpy array
#     show(m, new=True, axes=1).close() # optionally draw result on a new window


######################################################
print("Test 17")
def fibonacci_sphere(n):
    s = np.linspace(0, n, num=n, endpoint=False)
    theta = s * 2.399963229728653
    y = 1 - s * (2/(n-1))
    r = np.sqrt(1 - y * y)
    x = np.cos(theta) * r
    z = np.sin(theta) * r
    return [x,y,z]
# print(np.c_[fibonacci_sphere(10)].T)
fpoints = Points(np.c_[fibonacci_sphere(1000)].T)
if doshow:
    fpoints.show(axes=1).close()


######################################################
print("Test 18")
s = Sphere(res=10).linewidth(1).c("orange").compute_normals()
point_ids = s.labels('id', on="points").c('green')
cell_ids  = s.labels('id', on="cells").c('black')
if doshow:
    show(s, point_ids, cell_ids).close()



######################################################
print("Test 19")
sph = Sphere(quads=True, res=4).compute_normals().wireframe()
sph.celldata["zvals"] = sph.cell_centers().coordinates[:,2]
l2d = sph.labels("zvals", on="cells", precision=2).backcolor('orange9')
if doshow:
    show(sph, l2d, axes=1).close()



######################################################
print("Test 20")
c1 = Cube().rotate_z(5).x(2).y(1)
print("cube1 position", c1.pos())
T = c1.transform  # rotate by 5 degrees, sum 2 to x and 1 to y
c2 = Cube().c('r4')
c2.apply_transform(T)
c2.apply_transform(T)
c2.apply_transform(T)
print("cube2 position", c2.pos())
if doshow:
    show(c1, c2, axes=1).close()



######################################################
print("Test 21")
disc = Disc(r1=1, r2=1.2)
mesh = disc.extrude(3, res=50).linewidth(1)
mesh.cut_with_cylinder([0,0,2], r=0.4, axis='y', invert=True)
if doshow:
    show(mesh, axes=1)


######################################################
print("Test 22")
disc = Disc(r1=1, r2=1.2)
mesh = disc.extrude(3, res=50).linewidth(1)
mesh.cut_with_sphere([1,-0.7,2], r=1.5, invert=True)
if doshow:
    show(mesh, axes=1).close()


######################################################
print("Test 23")
arr = np.random.randn(100000, 3)/2
pts = Points(arr).c('red3').pos(5,0,0)
cube = Cube().pos(4,0.5,0)
assem = pts.cut_with_mesh(cube, keep=True)
if doshow:
    show(assem.unpack(), axes=1).close()


##################################################################### shapes.py
print("Test 24")
pts = [[1, 0, 0], [5, 2, 0], [3, 3, 1]]
ln = Line(pts, c='r', lw=5).pattern('- -', repeats=10)
if doshow:
    ln.show(axes=1).close()



######################################################
print("Test 25")
if doshow:
    shape = Assembly(dataurl+"timecourse1d.npy")[58]
    pts = shape.rotate_x(30).coordinates
    tangents = Line(pts).tangents()
    arrs = Arrows(pts, pts+tangents, c='blue9')
    show(shape.c('red5').lw(5), arrs, bg='bb', axes=1).close()


######################################################
print("Test 26")
if doshow:
    shape = Assembly(dataurl+"timecourse1d.npy")[55]
    curvs = Line(shape.coordinates).curvature()
    shape.cmap('coolwarm', curvs, vmin=-2,vmax=2).add_scalarbar3d(c='w')
    shape.render_lines_as_tubes().lw(12)
    pp = plot(curvs, ac='white', lc='yellow5')
    show(shape, pp, N=2, bg='bb', sharecam=False).close()



######################################################
print("Test 27")
aline = Line([(0,0,0),(1,3,0),(2,4,0)])
surf1 = aline.sweep((1,0.2,0), res=3)
surf2 = aline.sweep((0.2,0,1))
aline.color('r').linewidth(4)
if doshow:
    show(surf1, surf2, aline, axes=1).close()


######################################################
print("Test 28")
pts = [(-4,-3),(1,1),(2,4),(4,1),(3,-1),(2,-5),(9,-3)]
ln = Line(pts, c='r', lw=2).z(0.01)
rl = RoundedLine(pts, 0.6)
if doshow:
    show(Points(pts), ln, rl, axes=1).close()


######################################################
print("Test 29")
pts = np.random.randn(25,3)
for i,p in enumerate(pts):
    p += [5*i, 15*sin(i/2), i*i*i/200]
if doshow:
    show(Points(pts), Bezier(pts), axes=1).close()


######################################################
print("Test 30")
xcoords = np.arange(0, 2, 0.2)
ycoords = np.arange(0, 1, 0.2)
sqrtx = sqrt(xcoords)
grid = Grid(s=(sqrtx, ycoords))
if doshow:
    grid.show(axes=8)

# can also create a grid from np.mgrid:
X, Y = np.mgrid[-12:12:1000*1j, 0:15:1000*1j]
vgrid = Grid(s=(X[:,0], Y[0]))
if doshow:
    vgrid.show(axes=8).close()


######################################################
print("Test 31")
settings.immediate_rendering = False
plt = Plotter(N=18)
for i in range(18):
    ps = ParametricShape(i).color(i)
    if doshow:
        plt.at(i).show(ps, ps.name)
if doshow:
    plt.interactive()






# Example of usage of the madcad library
# See https://pymadcad.readthedocs.io/en/latest/index.html
import vedo
from madcad import *

##########################################################################
points = [O, X, X + Z, 2 * X + Z, 2 * (X + Z), X + 2 * Z, X + 5 * Z, 5 * Z]
section = Wire(points).segmented().flip() 
rev = revolution(2 * pi, (O, Z), section)
rev.mergeclose()
vedo.show("Revolution of a wire", rev, axes=7).close()


##########################################################################
# m = screw(10, 20)
# m["part"].option(color=vec3(70, 130, 180) / 255) # RGB
# vedo.show("A blue screw", m, axes=1).close()

##########################################################################
# Obtain two different shapes that has noting to to with each other
m1 = brick(width=vec3(2))
m2 = m1.transform(vec3(0.5, 0.3, 0.4)).transform(quat(0.7 * vec3(1, 1, 0)))
# Remove the volume of the second to the first
diff = difference(m1, m2)
vedo.show("Boolean difference", diff, axes=14).close()


##########################################################################
cube = brick(width=vec3(2))
bevel(
   cube,
   [(0, 1), (1, 2), (2, 3), (0, 3), (1, 5), (0, 4)],  # Edges to smooth
   ("width", 0.3),  # Cutting description, known as 'cutter'
)
vedo.show("A bevel cube", cube, axes=1).close()


##########################################################################
square_profile = square((O, Z), 5).flip()
primitives = [
    ArcCentered(( 5 * X,  Y),      O, 10 * X),
    ArcCentered((15 * X, -Y), 10 * X, 20 * X),
]
# Generate a path
path = web(primitives)
path.mergeclose()
m = tube(square_profile, path)

vmesh = vedo.utils.madcad2vedo(m)  # <-- convert to vedo.Mesh
print(vmesh)

scalar = vmesh.vertices[:, 0]
vmesh.cmap("rainbow", scalar, on="points").add_scalarbar(title="x-value")
vedo.show("Generating a path", vmesh, axes=7).close()

##########################################################################
c1 = Circle((vec3(0), Z), 1)
c2 = Circle((2 * X, X), 0.5)
c3 = (Circle((2 * Y, Y), 0.5), "tangent", 2)
e1 = extrusion(2 * Z, web(c1))

m = junction(e1, c2, c3, tangents="normal")
vm = vedo.utils.madcad2vedo(m)
vedo.show(vm, e1, axes=1, viewup="z").close()



from vedo import Volume, dataurl
from vedo.applications import Slicer2DPlotter

vol = Volume(dataurl+"embryo.slc")

plt = Slicer2DPlotter(vol)

# use cmap("bw") to have black and white color scale
# no argument will grab the existing cmap in vol (or use build_lut())
# plt.cmap('bw').lighting(window=120, level=25)

plt.show().close()
